"""

Copyright 2013 Steven Diamond, 2017 Robin Verschueren

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import unittest

import numpy as np
import scipy.sparse as sp
from scipy.linalg import lstsq

import cvxpy as cp
import cvxpy.tests.solver_test_helpers as sths
from cvxpy import Maximize, Minimize, Parameter, Problem
from cvxpy.atoms import (
    QuadForm,
    abs,
    huber,
    matrix_frac,
    norm,
    power,
    quad_over_lin,
    sum,
    sum_squares,
)
from cvxpy.expressions.variable import Variable
from cvxpy.reductions.solvers.defines import (
    INSTALLED_CONIC_SOLVERS,
    INSTALLED_SOLVERS,
    QP_SOLVERS,
    SOLVER_MAP_CONIC,
    SOLVER_MAP_QP,
)
from cvxpy.tests.base_test import BaseTest
from cvxpy.tests.solver_test_helpers import SolverTestHelper, StandardTestLPs, StandardTestQPs


class QPTestBase(BaseTest):
    """Base class with shared test helpers for QP-style problems."""

    def setUp(self) -> None:
        self.a = Variable(name='a')
        self.b = Variable(name='b')
        self.c = Variable(name='c')

        self.x = Variable(2, name='x')
        self.y = Variable(3, name='y')
        self.z = Variable(2, name='z')
        self.w = Variable(5, name='w')

        self.A = Variable((2, 2), name='A')
        self.B = Variable((2, 2), name='B')
        self.C = Variable((3, 2), name='C')

        self.slope = Variable(1, name='slope')
        self.offset = Variable(1, name='offset')
        self.quadratic_coeff = Variable(1, name='quadratic_coeff')

        T = 30
        self.position = Variable((2, T), name='position')
        self.velocity = Variable((2, T), name='velocity')
        self.force = Variable((2, T - 1), name='force')

        self.xs = Variable(80, name='xs')
        self.xsr = Variable(50, name='xsr')
        self.xef = Variable(80, name='xef')

    def solve_QP(self, problem, solver_name):
        """Override in subclasses."""
        raise NotImplementedError

    # License checking helpers - shared by subclasses
    @staticmethod
    def is_mosek_available():
        """Check if MOSEK is installed and a license is available."""
        if 'MOSEK' not in INSTALLED_SOLVERS:
            return False
        try:
            import mosek  # type: ignore
            env = mosek.Env()
            status = env.getlicense()
            return status == mosek.rescode.ok
        except Exception:
            return False

    @staticmethod
    def is_knitro_available():
        """Check if KNITRO is installed and a license is available."""
        if 'KNITRO' not in INSTALLED_SOLVERS:
            return False
        try:
            import knitro  # type: ignore
            kc = knitro.KN_new()
            if kc is None:
                return False
            knitro.KN_free(kc)
            return True
        except Exception:
            return False

    @staticmethod
    def is_xpress_available():
        """Check if XPRESS is installed and a license is available."""
        if 'XPRESS' not in INSTALLED_SOLVERS:
            return False
        try:
            import xpress  # type: ignore
            env = xpress.env()
            status = env.getlicense()
            return status == 0
        except Exception:
            return False

    def filter_licensed_solvers(self, solvers):
        """Remove solvers that don't have valid licenses."""
        result = list(solvers)
        if 'XPRESS' in result and not self.is_xpress_available():
            result.remove('XPRESS')
        if 'MOSEK' in result and not self.is_mosek_available():
            result.remove('MOSEK')
        if 'KNITRO' in result and not self.is_knitro_available():
            result.remove('KNITRO')
        return result

    def _check_kkt(self, problem, places=4):
        """Verify KKT conditions for a solved problem."""
        obj_pair = (problem.objective, None)
        var_pairs = [(v, None) for v in problem.variables()]
        con_pairs = [(c, None) for c in problem.constraints]
        sth = SolverTestHelper(obj_pair, var_pairs, con_pairs)
        # Problem already solved, just use the same problem object
        sth.prob = problem
        sth.check_primal_feasibility(places)
        sth.check_complementarity(places)
        sth.check_dual_domains(places)
        sth.check_stationary_lagrangian(places)

    # Test helper methods - shared by all subclasses
    def quad_over_lin(self, solver) -> None:
        p = Problem(Minimize(0.5 * quad_over_lin(abs(self.x-1), 1)),
                    [self.x <= -1])
        self.solve_QP(p, solver)
        for var in p.variables():
            self.assertItemsAlmostEqual(np.array([-1., -1.]),
                                        var.value, places=4)
        for con in p.constraints:
            self.assertItemsAlmostEqual(np.array([2., 2.]),
                                        con.dual_value, places=4)
        self._check_kkt(p, places=3)

    def abs(self, solver) -> None:
        u = Variable(2)
        constr = []
        constr += [abs(u[1] - u[0]) <= 100]
        prob = Problem(Minimize(sum_squares(u)), constr)
        print("The problem is QP: ", prob.is_qp())
        self.assertEqual(prob.is_qp(), True)
        result = prob.solve(solver=solver)
        self.assertAlmostEqual(result, 0)

    def power(self, solver) -> None:
        p = Problem(Minimize(sum(power(self.x, 2))), [])
        self.solve_QP(p, solver)
        for var in p.variables():
            self.assertItemsAlmostEqual([0., 0.], var.value, places=4)

    def power_matrix(self, solver) -> None:
        p = Problem(Minimize(sum(power(self.A - 3., 2))), [])
        self.solve_QP(p, solver)
        for var in p.variables():
            self.assertItemsAlmostEqual([3., 3., 3., 3.],
                                        var.value, places=4)

    def square_affine(self, solver) -> None:
        A = np.random.randn(10, 2)
        b = np.random.randn(10)
        p = Problem(Minimize(sum_squares(A @ self.x - b)))
        self.solve_QP(p, solver)
        for var in p.variables():
            self.assertItemsAlmostEqual(lstsq(A, b)[0].flatten(order='F'), var.value,
                                        places=1)

    def quad_form(self, solver) -> None:
        np.random.seed(0)
        A = np.random.randn(5, 5)
        z = np.random.randn(5)
        P = A.T.dot(A)
        q = -2*P.dot(z)
        p = Problem(Minimize(QuadForm(self.w, P) + q.T @ self.w))
        self.solve_QP(p, solver)
        for var in p.variables():
            self.assertItemsAlmostEqual(z, var.value, places=4)

    def rep_quad_form(self, solver) -> None:
        """A problem where the quad_form term is used multiple times.
        """
        np.random.seed(0)
        A = np.random.randn(5, 5)
        z = np.random.randn(5)
        P = A.T.dot(A)
        q = -2*P.dot(z)
        qf = QuadForm(self.w, P)
        p = Problem(Minimize(0.5*qf + 0.5*qf + q.T @ self.w))
        self.solve_QP(p, solver)
        for var in p.variables():
            self.assertItemsAlmostEqual(z, var.value, places=4)

    def affine_problem(self, solver) -> None:
        np.random.seed(0)
        A = np.random.randn(5, 2)
        A = np.maximum(A, 0)
        b = np.random.randn(5)
        b = np.maximum(b, 0)
        p = Problem(Minimize(sum(self.x)), [self.x >= 0, A @ self.x <= b])
        self.solve_QP(p, solver)
        for var in p.variables():
            self.assertItemsAlmostEqual([0., 0.], var.value, places=3)
        self._check_kkt(p, places=3)

    def maximize_problem(self, solver) -> None:
        np.random.seed(0)
        A = np.random.randn(5, 2)
        A = np.maximum(A, 0)
        b = np.random.randn(5)
        b = np.maximum(b, 0)
        p = Problem(Maximize(-sum(self.x)), [self.x >= 0, A @ self.x <= b])
        self.solve_QP(p, solver)
        for var in p.variables():
            self.assertItemsAlmostEqual([0., 0.], var.value, places=3)
        self._check_kkt(p, places=3)

    def norm_2(self, solver) -> None:
        A = np.random.randn(10, 5)
        b = np.random.randn(10)
        p = Problem(Minimize(norm(A @ self.w - b, 2)))
        self.solve_QP(p, solver)
        for var in p.variables():
            self.assertItemsAlmostEqual(lstsq(A, b)[0].flatten(order='F'), var.value,
                                        places=1)

    def mat_norm_2(self, solver) -> None:
        A = np.random.randn(5, 3)
        B = np.random.randn(5, 2)
        p = Problem(Minimize(norm(A @ self.C - B, 2)))
        s = self.solve_QP(p, solver)
        for var in p.variables():
            self.assertItemsAlmostEqual(lstsq(A, B)[0],
                                        s.primal_vars[var.id], places=1)

    def quad_form_coeff(self, solver) -> None:
        np.random.seed(0)
        A = np.random.randn(5, 5)
        z = np.random.randn(5)
        P = A.T.dot(A)
        q = -2*P.dot(z)
        p = Problem(Minimize(QuadForm(self.w, P) + q.T @ self.w))
        self.solve_QP(p, solver)
        for var in p.variables():
            self.assertItemsAlmostEqual(z, var.value, places=4)

    def quad_form_bound(self, solver) -> None:
        P = np.array([[13, 12, -2], [12, 17, 6], [-2, 6, 12]])
        q = np.array([[-22], [-14.5], [13]])
        r = 1
        y_star = np.array([[1], [0.5], [-1]])
        p = Problem(Minimize(0.5*QuadForm(self.y, P) + q.T @ self.y + r),
                    [self.y >= -1, self.y <= 1])
        self.solve_QP(p, solver)
        for var in p.variables():
            self.assertItemsAlmostEqual(y_star, var.value, places=4)
        self._check_kkt(p)

    def regression_1(self, solver) -> None:
        np.random.seed(1)
        # Number of examples to use
        n = 100
        # Specify the true value of the variable
        true_coeffs = np.array([[2, -2, 0.5]]).T
        # Generate data
        x_data = np.random.rand(n) * 5
        x_data = np.atleast_2d(x_data)
        x_data_expanded = np.vstack([np.power(x_data, i)
                                     for i in range(1, 4)])
        x_data_expanded = np.atleast_2d(x_data_expanded)
        y_data = x_data_expanded.T.dot(true_coeffs) + 0.5 * np.random.rand(n, 1)
        y_data = np.atleast_2d(y_data)

        line = self.offset + x_data * self.slope
        residuals = line.T - y_data
        fit_error = sum_squares(residuals)
        p = Problem(Minimize(fit_error), [])
        self.solve_QP(p, solver)
        self.assertAlmostEqual(1171.60037715, p.value, places=4)

    def regression_2(self, solver) -> None:
        np.random.seed(1)
        # Number of examples to use
        n = 100
        # Specify the true value of the variable
        true_coeffs = np.array([2, -2, 0.5])
        # Generate data
        x_data = np.random.rand(n) * 5
        x_data_expanded = np.vstack([np.power(x_data, i)
                                     for i in range(1, 4)])
        print(x_data_expanded.shape, true_coeffs.shape)
        y_data = x_data_expanded.T.dot(true_coeffs) + 0.5 * np.random.rand(n)

        quadratic = self.offset + x_data * self.slope + \
            self.quadratic_coeff*np.power(x_data, 2)
        residuals = quadratic.T - y_data
        fit_error = sum_squares(residuals)
        p = Problem(Minimize(fit_error), [])
        self.solve_QP(p, solver)

        self.assertAlmostEqual(139.225660756, p.value, places=4)

    def control(self, solver) -> None:
        # Some constraints on our motion
        # The object should start from the origin, and end at rest
        initial_velocity = np.array([-20, 100])
        final_position = np.array([100, 100])
        T = 30  # The number of timesteps
        h = 0.1  # The time between time intervals
        mass = 1  # Mass of object
        drag = 0.1  # Drag on object
        g = np.array([0, -9.8])  # Gravity on object
        # Create a problem instance
        constraints = []
        # Add constraints on our variables
        for i in range(T - 1):
            constraints += [self.position[:, i + 1] == self.position[:, i] +
                            h * self.velocity[:, i]]
            acceleration = self.force[:, i]/mass + g - \
                drag * self.velocity[:, i]
            constraints += [self.velocity[:, i + 1] == self.velocity[:, i] +
                            h * acceleration]

        # Add position constraints
        constraints += [self.position[:, 0] == 0]
        constraints += [self.position[:, -1] == final_position]
        # Add velocity constraints
        constraints += [self.velocity[:, 0] == initial_velocity]
        constraints += [self.velocity[:, -1] == 0]
        # Solve the problem
        p = Problem(Minimize(.01 * sum_squares(self.force)), constraints)
        self.solve_QP(p, solver)
        self.assertAlmostEqual(1059.616, p.value, places=1)
        # KKT check skipped: check_stationary_lagrangian fails for 2D matrix
        # variables due to inconsistent gradient ordering (sum_squares uses
        # C order, constraint terms use F order). TODO fix this
        # self._check_kkt(p, places=3)

    def sparse_system(self, solver) -> None:
        m = 100
        n = 80
        np.random.seed(1)
        density = 0.4
        A = sp.random_array((m, n), density=density)
        b = np.random.randn(m)

        p = Problem(Minimize(sum_squares(A @ self.xs - b)), [self.xs == 0])
        self.solve_QP(p, solver)
        self.assertAlmostEqual(b.T.dot(b), p.value, places=4)

    def smooth_ridge(self, solver) -> None:
        np.random.seed(1)
        n = 50
        k = 20
        eta = 1

        A = np.ones((k, n))
        b = np.ones((k))
        obj = sum_squares(A @ self.xsr - b) + \
            eta*sum_squares(self.xsr[:-1]-self.xsr[1:])
        p = Problem(Minimize(obj), [])
        self.solve_QP(p, solver)
        self.assertAlmostEqual(0, p.value, places=4)

    def huber_small(self, solver, places=4) -> None:
        # Solve the Huber regression problem
        x = Variable(3)
        objective = sum(huber(x))

        # Solve problem with QP
        p = Problem(Minimize(objective), [x[2] >= 3])
        self.solve_QP(p, solver)
        self.assertAlmostEqual(3, x.value[2], places=places)
        self.assertAlmostEqual(5, objective.value, places=places)
        self._check_kkt(p, places=places)

    def huber(self, solver) -> None:
        # Generate problem data
        n = 3
        m = 5
        data = [0.89, 0.39, 0.96, 0.34, 0.68, 0.18, 0.63, 0.42, 0.51, 0.66, 0.43, 0.77]
        indices = [0, 1, 2, 3, 4, 2, 3, 0, 1, 2, 3, 4]
        indptr = [0, 5, 7, 12]
        A = sp.csc_array((data, indices, indptr), shape=(m, n))
        x_true = np.random.randn(n) / np.sqrt(n)
        ind95 = (np.random.rand(m) < 0.95).astype(float)
        b = A.dot(x_true) + np.multiply(0.5*np.random.randn(m), ind95) \
            + np.multiply(10.*np.random.rand(m), 1. - ind95)

        # Solve the Huber regression problem
        x = Variable(n)
        objective = sum(huber(A @ x - b))

        # Solve problem with QP
        p = Problem(Minimize(objective))
        self.solve_QP(p, solver)
        self.assertAlmostEqual(1.452797819667, objective.value, places=3)
        self.assertItemsAlmostEqual(x.value,
                                    [1.20524645, -0.85271489, -0.50838494],
                                    places=3)

    def equivalent_forms_1(self, solver) -> None:
        m = 100
        n = 80
        r = 70
        np.random.seed(1)
        A = np.random.randn(m, n)
        b = np.random.randn(m)
        G = np.random.randn(r, n)
        h = np.random.randn(r)

        obj1 = .1 * sum((A @ self.xef - b) ** 2)
        cons = [G @ self.xef == h]

        p1 = Problem(Minimize(obj1), cons)
        self.solve_QP(p1, solver)
        self.assertAlmostEqual(p1.value, 68.1119420108, places=4)
        self._check_kkt(p1, places=4)

    def equivalent_forms_2(self, solver) -> None:
        m = 100
        n = 80
        r = 70
        np.random.seed(1)
        A = np.random.randn(m, n)
        b = np.random.randn(m)
        G = np.random.randn(r, n)
        h = np.random.randn(r)

        # ||Ax-b||^2 = x^T (A^T A) x - 2(A^T b)^T x + ||b||^2
        P = np.dot(A.T, A)
        q = -2*np.dot(A.T, b)
        r = np.dot(b.T, b)

        obj2 = .1*(QuadForm(self.xef, P)+q.T @ self.xef+r)
        cons = [G @ self.xef == h]

        p2 = Problem(Minimize(obj2), cons)
        self.solve_QP(p2, solver)
        self.assertAlmostEqual(p2.value, 68.1119420108, places=4)
        self._check_kkt(p2, places=4)

    def equivalent_forms_3(self, solver) -> None:
        m = 100
        n = 80
        r = 70
        np.random.seed(1)
        A = np.random.randn(m, n)
        b = np.random.randn(m)
        G = np.random.randn(r, n)
        h = np.random.randn(r)

        # ||Ax-b||^2 = x^T (A^T A) x - 2(A^T b)^T x + ||b||^2
        P = np.dot(A.T, A)
        q = -2*np.dot(A.T, b)
        r = np.dot(b.T, b)
        Pinv = np.linalg.inv(P)

        obj3 = .1 * (matrix_frac(self.xef, Pinv)+q.T @ self.xef+r)
        cons = [G @ self.xef == h]

        p3 = Problem(Minimize(obj3), cons)
        self.solve_QP(p3, solver)
        print(solver)
        self.assertAlmostEqual(p3.value, 68.1119420108, places=4)
        self._check_kkt(p3, places=4)


class TestQp(QPTestBase):
    """Test native QP solvers."""

    def setUp(self) -> None:
        super().setUp()

        # Check for all installed QP solvers
        self.solvers = [x for x in QP_SOLVERS if x in INSTALLED_SOLVERS]
        self.solvers = self.filter_licensed_solvers(self.solvers)

    def solve_QP(self, problem, solver_name):
        return problem.solve(solver=solver_name, verbose=False)

    def test_all_solvers(self) -> None:
        for solver in self.solvers:
            self.quad_over_lin(solver)
            self.power(solver)
            self.power_matrix(solver)
            self.square_affine(solver)
            self.quad_form(solver)
            self.affine_problem(solver)
            self.maximize_problem(solver)
            self.abs(solver)

            # Do we need the following functionality?
            # self.norm_2(solver)
            # self.mat_norm_2(solver)

            self.quad_form_coeff(solver)
            self.quad_form_bound(solver)
            self.regression_1(solver)
            self.regression_2(solver)
            self.rep_quad_form(solver)

            # slow tests:
            self.control(solver)
            self.sparse_system(solver)
            self.smooth_ridge(solver)
            self.huber_small(solver)
            self.huber(solver)
            self.equivalent_forms_1(solver)
            self.equivalent_forms_2(solver)
            if solver != cp.KNITRO:
                self.equivalent_forms_3(solver)

    def test_qp_bound_attr(self) -> None:
        for solver in self.solvers:
            solver_cls = SOLVER_MAP_QP.get(solver)
            if solver_cls is not None and getattr(solver_cls, 'BOUNDED_VARIABLES', False):
                StandardTestQPs.test_qp_bound_attr(solver=solver)

    def test_warm_start(self) -> None:
        """Test warm start.
        """
        m = 200
        n = 100
        np.random.seed(1)
        A = np.random.randn(m, n)
        b = Parameter(m)

        # Construct the problem.
        x = Variable(n)
        prob = Problem(Minimize(sum_squares(A @ x - b)))

        b.value = np.random.randn(m)
        result = prob.solve(solver="OSQP", warm_start=False)
        result2 = prob.solve(solver="OSQP", warm_start=True)
        self.assertAlmostEqual(result, result2)
        b.value = np.random.randn(m)
        result = prob.solve(solver="OSQP", warm_start=True)
        result2 = prob.solve(solver="OSQP", warm_start=False)
        self.assertAlmostEqual(result, result2)

    def test_qpalm_warmstart(self) -> None:
        """Test warm start.
        """
        if cp.QPALM in INSTALLED_SOLVERS:
            m = 200
            n = 100
            np.random.seed(1)
            A = np.random.randn(m, n)
            b = Parameter(m)

            # Construct the problem.
            x = Variable(n)
            prob = Problem(Minimize(sum_squares(A @ x - b)))

            b.value = np.random.randn(m)
            result = prob.solve(solver=cp.QPALM, warm_start=False)
            result2 = prob.solve(solver=cp.QPALM, warm_start=True)
            self.assertAlmostEqual(result, result2)
            b.value = np.random.randn(m)
            result = prob.solve(solver=cp.QPALM, warm_start=True)
            result2 = prob.solve(solver=cp.QPALM, warm_start=False)
            self.assertAlmostEqual(result, result2)

    def test_gurobi_warmstart(self) -> None:
        """Test Gurobi warm start with a user provided point.
        """
        if cp.GUROBI in INSTALLED_SOLVERS:
            import gurobipy
            m = 4
            n = 3

            y = Variable(nonneg=True)
            X = Variable((m, n))
            X_vals = np.reshape(np.arange(m*n), (m, n))
            prob = Problem(Minimize(y**2 + cp.sum(X)), [X == X_vals])
            X.value = X_vals + 1
            prob.solve(solver=cp.GUROBI, warm_start=True)
            # Check that "start" value was set appropriately.
            model = prob.solver_stats.extra_stats
            model_x = model.getVars()
            assert gurobipy.GRB.UNDEFINED == model_x[0].start
            assert np.isclose(0, model_x[0].x)
            for i in range(1, X.size + 1):
                row = (i - 1) % X.shape[0]
                col = (i - 1) // X.shape[0]
                assert X_vals[row, col] + 1 == model_x[i].start
                assert np.isclose(X.value[row, col], model_x[i].x)

    def test_xpress_warmstart(self) -> None:
        """Test XPRESS warm start with a user provided point.
        """
        if cp.XPRESS in INSTALLED_SOLVERS:
            m = 20
            n = 10
            np.random.seed(1)
            A = np.random.randn(m, n)
            b = Parameter(m)

            # Construct the problem.
            x = Variable(n, integer=True)
            prob = Problem(Minimize(sum_squares(A @ x - b)))

            b.value = np.random.randn(m)
            result = prob.solve(solver=cp.XPRESS, warm_start=False)
            result2 = prob.solve(solver=cp.XPRESS, warm_start=True)
            self.assertAlmostEqual(result, result2)
            x.value = x.value.astype(np.int64)

            xprime = Variable(n, integer=True)
            prob = Problem(Minimize(sum_squares(A @ xprime - b)))
            xprime.value = x.value
            result = prob.solve(solver=cp.XPRESS, warm_start=True)
            result2 = prob.solve(solver=cp.XPRESS, warm_start=False)
            self.assertAlmostEqual(result, result2)

    def test_highs_warmstart(self) -> None:
        """Test warm start.
        """
        if cp.HIGHS in INSTALLED_SOLVERS:
            m = 200
            n = 100
            np.random.seed(1)
            A = np.random.randn(m, n)
            b = Parameter(m)

            # Construct the problem.
            x = Variable(n)
            prob = Problem(Minimize(sum_squares(A @ x - b)))

            b.value = np.random.randn(m)
            result = prob.solve(solver=cp.HIGHS, warm_start=False)
            result2 = prob.solve(solver=cp.HIGHS, warm_start=True)
            self.assertAlmostEqual(result, result2)
            b.value = np.random.randn(m)
            result = prob.solve(solver=cp.HIGHS, warm_start=True)
            result2 = prob.solve(solver=cp.HIGHS, warm_start=False)
            self.assertAlmostEqual(result, result2)

    def test_highs_cvar(self) -> None:
        """Test problem with CVaR constraint from
        https://github.com/cvxpy/cvxpy/issues/2836
        """
        if cp.HIGHS in INSTALLED_SOLVERS:
            # Generate data
            num_stocks = 5
            num_samples = 25
            np.random.seed(1)
            pnl_samples = np.random.uniform(low=0.0, high=1.0, size=(num_samples, num_stocks))
            pnl_expected = pnl_samples.mean(axis=0)

            # Prepare to solve
            quantile = 0.05
            w = cp.Variable(num_stocks, nonneg=True)
            cvar = cp.cvar(pnl_samples @ w, 1 - quantile)
            pnl = w @ pnl_expected

            # Solve
            objective = cp.Maximize(pnl)
            constraints = [cvar <= 0.5]
            problem = cp.Problem(objective, constraints)
            problem.solve(
                solver=cp.HIGHS,
            )
            assert problem.status == cp.OPTIMAL

    def test_piqp_warmstart(self) -> None:
        """Test warm start.
        """
        if cp.PIQP in INSTALLED_SOLVERS:
            m = 200
            n = 100
            np.random.seed(1)
            A = np.random.randn(m, n)
            b = Parameter(m)

            # Construct the problem.
            x = Variable(n)
            prob = Problem(Minimize(sum_squares(A @ x - b)))

            b.value = np.random.randn(m)
            result = prob.solve(solver=cp.PIQP, warm_start=False)
            result2 = prob.solve(solver=cp.PIQP, warm_start=True)
            self.assertAlmostEqual(result, result2)
            b.value = np.random.randn(m)
            result = prob.solve(solver=cp.PIQP, warm_start=True)
            result2 = prob.solve(solver=cp.PIQP, warm_start=False)
            self.assertAlmostEqual(result, result2)
    
    def test_copt_warmstart(self) -> None:
        """Test warm start.
        """
        if cp.COPT in INSTALLED_SOLVERS:
            m = 200
            n = 100
            np.random.seed(1)
            A = np.random.randn(m, n)
            b = Parameter(m)
            
            # Construct the problem.
            x = Variable(n)
            prob = Problem(Minimize(sum_squares(A @ x - b)))

            b.value = np.random.randn(m)
            result = prob.solve(solver=cp.COPT, warm_start=False)
            result2 = prob.solve(solver=cp.COPT, warm_start=True)
            self.assertAlmostEqual(result, result2)
            b.value = np.random.randn(m)
            result = prob.solve(solver=cp.COPT, warm_start=True)
            result2 = prob.solve(solver=cp.COPT, warm_start=False)
            self.assertAlmostEqual(result, result2)

    def test_parametric(self) -> None:
        """Test solve parametric problem vs full problem"""
        x = Variable()
        a = 10
        #  b_vec = [-10, -2., 2., 3., 10.]
        b_vec = [-10, -2.]

        for solver in self.solvers:

            print(solver)
            # Solve from scratch with no parameters
            x_full = []
            obj_full = []
            for b in b_vec:
                obj = Minimize(a * (x ** 2) + b * x)
                constraints = [0 <= x, x <= 1]
                prob = Problem(obj, constraints)
                prob.solve(solver=solver)
                x_full += [x.value]
                obj_full += [prob.value]

            # Solve parametric
            x_param = []
            obj_param = []
            b = Parameter()
            obj = Minimize(a * (x ** 2) + b * x)
            constraints = [0 <= x, x <= 1]
            prob = Problem(obj, constraints)
            for b_value in b_vec:
                b.value = b_value
                prob.solve(solver=solver)
                x_param += [x.value]
                obj_param += [prob.value]

            print(x_full)
            print(x_param)
            for i in range(len(b_vec)):
                self.assertItemsAlmostEqual(x_full[i], x_param[i], places=3)
                self.assertAlmostEqual(obj_full[i], obj_param[i])

    def test_square_param(self) -> None:
        """Test issue arising with square plus parameter.
        """
        a = Parameter(value=1)
        b = Variable()

        obj = Minimize(b ** 2 + abs(a))
        prob = Problem(obj)
        prob.solve(solver="SCS")
        self.assertAlmostEqual(obj.value, 1.0)

    def test_gurobi_time_limit_no_solution(self) -> None:
        """Make sure that if Gurobi terminates due to a time limit before finding a solution:
            1) no error is raised,
            2) solver stats are returned.
            The test is skipped if something changes on Gurobi's side so that:
            - a solution is found despite a time limit of zero,
            - a different termination criteria is hit first.
        """
        from cvxpy import GUROBI
        if GUROBI in INSTALLED_SOLVERS:
            import gurobipy
            objective = Minimize(self.x[0])
            constraints = [self.x[0] >= 1]
            prob = Problem(objective, constraints)
            try:
                prob.solve(solver=GUROBI, TimeLimit=0.0)
            except Exception as e:
                self.fail("An exception %s is raised instead of returning a result." % e)

            extra_stats = None
            solver_stats = getattr(prob, "solver_stats", None)
            if solver_stats:
                extra_stats = getattr(solver_stats, "extra_stats", None)
            self.assertTrue(extra_stats, "Solver stats have not been returned.")

            nb_solutions = getattr(extra_stats, "SolCount", None)
            if nb_solutions:
                self.skipTest("Gurobi has found a solution, the test is not relevant anymore.")

            solver_status = getattr(extra_stats, "Status", None)
            if solver_status != gurobipy.GRB.TIME_LIMIT:
                self.skipTest("Gurobi terminated for a different reason than reaching time limit, "
                              "the test is not relevant anymore.")

        else:
            with self.assertRaises(Exception) as cm:
                prob = Problem(Minimize(norm(self.x, 1)), [self.x == 0])
                prob.solve(solver=GUROBI, TimeLimit=0)
            self.assertEqual(str(cm.exception), "The solver %s is not installed." % GUROBI)

    def test_gurobi_environment(self) -> None:
        """Tests that Gurobi environments can be passed to Model.
        Gurobi environments can include licensing and model parameter data.
        """
        from cvxpy import GUROBI
        if GUROBI in INSTALLED_SOLVERS:
            import gurobipy

            # Set a few parameters to random values close to their defaults
            params = {
                'MIPGap': np.random.random(),  # range {0, INFINITY}
                'AggFill': np.random.randint(10),  # range {-1, MAXINT}
                'PerturbValue': np.random.random(),  # range: {0, INFINITY}
            }

            # Create a custom environment and set some parameters
            custom_env = gurobipy.Env()
            for k, v in params.items():
                custom_env.setParam(k, v)

            # Testing QP Solver Interface
            sth = StandardTestLPs.test_lp_0(solver='GUROBI', env=custom_env)
            model = sth.prob.solver_stats.extra_stats
            for k, v in params.items():
                # https://www.gurobi.com/documentation/9.1/refman/py_model_getparaminfo.html
                name, p_type, p_val, p_min, p_max, p_def = model.getParamInfo(k)
                self.assertEqual(v, p_val)

        else:
            with self.assertRaises(Exception) as cm:
                prob = Problem(Minimize(norm(self.x, 1)), [self.x == 0])
                prob.solve(solver=GUROBI, TimeLimit=0)
            self.assertEqual(str(cm.exception), "The solver %s is not installed." % GUROBI)


class TestConicQuadObj(QPTestBase):
    """Test conic solvers with use_quad_obj=True."""

    def setUp(self) -> None:
        super().setUp()

        # Conic solvers that support quadratic objectives
        # Exclude KNITRO - its conic interface with use_quad_obj is unstable in CI
        self.solvers = [
            solver for solver in INSTALLED_CONIC_SOLVERS
            if solver in SOLVER_MAP_CONIC
            and SOLVER_MAP_CONIC[solver].supports_quad_obj()
            and solver != cp.KNITRO
        ]
        self.solvers = self.filter_licensed_solvers(self.solvers)

    def solve_QP(self, problem, solver_name):
        """Solve with use_quad_obj=True and verify no SOC cones are introduced."""
        data, _, _ = problem.get_problem_data(
            solver_name,
            solver_opts={"use_quad_obj": True}
        )
        self.assertEqual(data["dims"].soc, [],
            f"Problem should have no SOC cones for QP canonicalization with {solver_name}")
        return problem.solve(solver=solver_name, use_quad_obj=True, verbose=False)

    def test_all_solvers(self) -> None:
        """Test conic solvers with use_quad_obj=True.

        Only runs tests that have constraints (filtering out unconstrained
        problems for solvers with REQUIRES_CONSTR=True).

        Tests with m=0 after canonicalization (skip for REQUIRES_CONSTR solvers):
        - power, quad_form, quad_form_coeff, rep_quad_form
        """
        for solver in self.solvers:
            requires_constr = SOLVER_MAP_CONIC[solver].REQUIRES_CONSTR

            self.quad_over_lin(solver)
            if not requires_constr:
                self.power(solver)
            self.power_matrix(solver)
            self.square_affine(solver)
            if not requires_constr:
                self.quad_form(solver)
            self.affine_problem(solver)
            self.maximize_problem(solver)
            self.abs(solver)
            if not requires_constr:
                self.quad_form_coeff(solver)
            self.quad_form_bound(solver)
            self.regression_1(solver)
            self.regression_2(solver)
            if not requires_constr:
                self.rep_quad_form(solver)
            self.control(solver)
            self.sparse_system(solver)
            self.smooth_ridge(solver)
            self.huber_small(solver, places=3)
            self.huber(solver)
            self.equivalent_forms_1(solver)
            self.equivalent_forms_2(solver)
            self.equivalent_forms_3(solver)


@unittest.skipUnless('MPAX' in INSTALLED_SOLVERS, 'MPAX is not installed.')
class TestMPAX(unittest.TestCase):

    def test_mpax_lp_0(self) -> None:
        StandardTestLPs.test_lp_0(solver='MPAX')

    def test_mpax_lp_1(self) -> None:
        StandardTestLPs.test_lp_1(solver='MPAX')

    def test_mpax_lp_2(self) -> None:
        StandardTestLPs.test_lp_2(solver='MPAX')

    def test_mpax_lp_3(self) -> None:
        sth = sths.lp_3()
        with self.assertWarns(Warning):
            sth.prob.solve(solver='MPAX')
            self.assertEqual(sth.prob.status, cp.settings.INFEASIBLE_OR_UNBOUNDED)

    def test_mpax_lp_4(self) -> None:
            sth = sths.lp_4()
            with self.assertWarns(Warning):
                sth.prob.solve(solver='MPAX')
                self.assertEqual(sth.prob.status, cp.settings.INFEASIBLE_OR_UNBOUNDED)

    def test_mpax_lp_5(self) -> None:
        StandardTestLPs.test_lp_5(solver='MPAX')

    def test_mpax_lp_6(self) -> None:
        StandardTestLPs.test_lp_6(solver='MPAX')

    def test_mpax_warmstart(self) -> None:
        x = cp.Variable(shape=(2,), name='x')
        objective = cp.Minimize(-4 * x[0] - 5 * x[1])
        constraints = [2 * x[0] + x[1] <= 3,
                    x[0] + 2 * x[1] <= 3,
                    x[0] >= 0,
                    x[1] >= 0]
        prob = cp.Problem(objective, constraints)
        result1 = prob.solve(solver='MPAX', warm_start=False)
        self.assertAlmostEqual(result1, -9, places=4)
        result2 = prob.solve(solver='MPAX', warm_start=True)
        self.assertAlmostEqual(result2, -9, places=4)

    def test_MPAX_qp_0(self) -> None:
        StandardTestQPs.test_qp_0(solver='MPAX')


class TestQpSolverValidation(unittest.TestCase):
    """Test QP solver validation of unsupported cone types."""

    def _apply_reductions(self, problem):
        """Apply the full reduction chain to get a ParamConeProg."""
        from cvxpy.reductions.cvx_attr2constr import CvxAttr2Constr
        from cvxpy.reductions.dcp2cone.cone_matrix_stuffing import ConeMatrixStuffing
        from cvxpy.reductions.dcp2cone.dcp2cone import Dcp2Cone
        from cvxpy.reductions.flip_objective import FlipObjective

        reductions = []
        if isinstance(problem.objective, cp.Maximize):
            reductions.append(FlipObjective())
        reductions.extend([Dcp2Cone(), CvxAttr2Constr(), ConeMatrixStuffing()])

        reduced = problem
        for reduction in reductions:
            reduced = reduction.apply(reduced)[0]
        return reduced

    def test_qp_solver_rejects_exponential_cones(self) -> None:
        """Test that QP solver rejects problems with exponential cones."""
        from cvxpy.error import SolverError
        from cvxpy.reductions.solvers.qp_solvers.osqp_qpif import OSQP

        # Create a problem with exponential cone (log)
        x = cp.Variable()
        prob = cp.Problem(cp.Maximize(cp.log(x)), [x <= 1])

        # Apply reductions to get ParamConeProg
        param_cone_prog = self._apply_reductions(prob)

        # Test accepts() returns False
        osqp_solver = OSQP()
        self.assertFalse(osqp_solver.accepts(param_cone_prog))

        # Test apply() raises SolverError with helpful message
        with self.assertRaises(SolverError) as cm:
            osqp_solver.apply(param_cone_prog)
        self.assertIn("exponential cones", str(cm.exception))
        self.assertIn("OSQP", str(cm.exception))

    def test_qp_solver_rejects_psd_cones(self) -> None:
        """Test that QP solver rejects problems with PSD cones."""
        from cvxpy.error import SolverError
        from cvxpy.reductions.solvers.qp_solvers.osqp_qpif import OSQP

        # Create a problem with PSD cone
        X = cp.Variable((2, 2), symmetric=True)
        prob = cp.Problem(cp.Minimize(cp.trace(X)), [X >> 0, X[0, 0] >= 1])

        # Apply reductions to get ParamConeProg
        param_cone_prog = self._apply_reductions(prob)

        # Test accepts() returns False
        osqp_solver = OSQP()
        self.assertFalse(osqp_solver.accepts(param_cone_prog))

        # Test apply() raises SolverError with helpful message
        with self.assertRaises(SolverError) as cm:
            osqp_solver.apply(param_cone_prog)
        self.assertIn("PSD cones", str(cm.exception))

    def test_qp_solver_rejects_soc_cones(self) -> None:
        """Test that QP solver rejects problems with second-order cones."""
        from cvxpy.error import SolverError
        from cvxpy.reductions.solvers.qp_solvers.osqp_qpif import OSQP

        # Create a problem with SOC (norm)
        x = cp.Variable(3)
        prob = cp.Problem(cp.Minimize(cp.norm(x)), [cp.sum(x) == 1])

        # Apply reductions to get ParamConeProg
        param_cone_prog = self._apply_reductions(prob)

        # Test accepts() returns False
        osqp_solver = OSQP()
        self.assertFalse(osqp_solver.accepts(param_cone_prog))

        # Test apply() raises SolverError with helpful message
        with self.assertRaises(SolverError) as cm:
            osqp_solver.apply(param_cone_prog)
        self.assertIn("second-order cones", str(cm.exception))



"""
Copyright 2025 CVXPY developers

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
from scipy import sparse

import cvxpy.settings as s
from cvxpy.reductions.solution import Solution, failure_solution
from cvxpy.reductions.solvers import utilities
from cvxpy.reductions.solvers.qp_solvers.qp_solver import QpSolver
from cvxpy.utilities.citations import CITATION_DICT


def kn_isinf(x) -> bool:
    """Check if x is -inf or inf."""
    if x <= -np.inf or x >= np.inf:
        return True
    if x <= float("-inf") or x >= float("inf"):
        return True

    import knitro as kn

    if x <= -kn.KN_INFINITY or x >= kn.KN_INFINITY:
        return True
    return False


def kn_rm_inf(a) -> tuple[list[int], list[float]]:
    """Convert -inf to -kn.KN_INFINITY and inf to kn.KN_INFINITY."""
    i, vs = [], []
    for j, v in enumerate(a):
        if not kn_isinf(v):
            i.append(j)
            vs.append(v)
    return i, vs


class KNITRO(QpSolver):
    """QP interface for the Knitro solver"""

    MIP_CAPABLE = True
    BOUNDED_VARIABLES = True

    # Keys:
    CONTEXT_KEY = "context"
    X_INIT_KEY = "x_init"
    Y_INIT_KEY = "y_init"

    # Keyword arguments for the CVXPY interface.
    INTERFACE_ARGS = [X_INIT_KEY, Y_INIT_KEY]

    # Map of Knitro status to CVXPY status.
    # This is based on the Knitro documentation:
    # https://www.artelys.com/app/docs/knitro/3_referenceManual/returnCodes.html
    STATUS_MAP = {
        0: s.OPTIMAL,
        -100: s.OPTIMAL_INACCURATE,
        -101: s.USER_LIMIT,
        -102: s.USER_LIMIT,
        -103: s.USER_LIMIT,
        -200: s.INFEASIBLE,
        -201: s.INFEASIBLE,
        -202: s.INFEASIBLE,
        -203: s.INFEASIBLE,
        -204: s.INFEASIBLE,
        -205: s.INFEASIBLE,
        -300: s.UNBOUNDED,
        -301: s.UNBOUNDED,
        -400: s.USER_LIMIT,
        -401: s.USER_LIMIT,
        -402: s.USER_LIMIT,
        -403: s.USER_LIMIT,
        -404: s.USER_LIMIT,
        -405: s.USER_LIMIT,
        -406: s.USER_LIMIT,
        -410: s.USER_LIMIT,
        -411: s.USER_LIMIT,
        -412: s.USER_LIMIT,
        -413: s.USER_LIMIT,
        -415: s.USER_LIMIT,
        -416: s.USER_LIMIT,
        -500: s.SOLVER_ERROR,
        -501: s.SOLVER_ERROR,
        -502: s.SOLVER_ERROR,
        -503: s.SOLVER_ERROR,
        -504: s.SOLVER_ERROR,
        -505: s.SOLVER_ERROR,
        -506: s.SOLVER_ERROR,
        -507: s.SOLVER_ERROR,
        -508: s.SOLVER_ERROR,
        -509: s.SOLVER_ERROR,
        -510: s.SOLVER_ERROR,
        -511: s.SOLVER_ERROR,
        -512: s.SOLVER_ERROR,
        -513: s.SOLVER_ERROR,
        -514: s.SOLVER_ERROR,
        -515: s.SOLVER_ERROR,
        -516: s.SOLVER_ERROR,
        -517: s.SOLVER_ERROR,
        -518: s.SOLVER_ERROR,
        -519: s.SOLVER_ERROR,
        -520: s.SOLVER_ERROR,
        -521: s.SOLVER_ERROR,
        -522: s.SOLVER_ERROR,
        -523: s.SOLVER_ERROR,
        -524: s.SOLVER_ERROR,
        -525: s.SOLVER_ERROR,
        -526: s.SOLVER_ERROR,
        -527: s.SOLVER_ERROR,
        -528: s.SOLVER_ERROR,
        -529: s.SOLVER_ERROR,
        -530: s.SOLVER_ERROR,
        -531: s.SOLVER_ERROR,
        -532: s.SOLVER_ERROR,
        -600: s.SOLVER_ERROR,
    }

    def name(self):
        return s.KNITRO

    def import_solver(self) -> None:
        """Imports the Knitro solver."""
        import knitro  # noqa: F401

    def apply(self, problem):
        """
        Construct QP problem data stored in a dictionary.
        The QP has the following form

            minimize      1/2 x' P x + q' x
            subject to    A x =  b
                          F x <= g

        """
        return super(KNITRO, self).apply(problem)

    def invert(self, results, inverse_data):
        import knitro as kn

        if KNITRO.CONTEXT_KEY not in results:
            return failure_solution(s.SOLVER_ERROR)

        kc = results[KNITRO.CONTEXT_KEY]
        num_iters = kn.KN_get_number_iters(kc)
        solve_time = kn.KN_get_solve_time_real(kc)
        attr = {
            s.SOLVE_TIME: solve_time,
            s.NUM_ITERS: num_iters,
            s.EXTRA_STATS: kc,
        }
        if s.STATUS in results and results[s.STATUS] == s.SOLVER_ERROR:
            solution = failure_solution(s.SOLVER_ERROR, attr)
        else:
            status_kn, obj_kn, x_kn, y_kn = kn.KN_get_solution(kc)
            status = self.STATUS_MAP.get(status_kn, s.SOLVER_ERROR)
            if status == s.UNBOUNDED:
                solution = Solution(status, -np.inf, {}, {}, attr)
            elif (status not in s.SOLUTION_PRESENT) or (x_kn is None):
                solution = failure_solution(status, attr)
            else:
                obj = obj_kn + inverse_data[s.OFFSET]
                x = np.array(x_kn)
                primal_vars = {KNITRO.VAR_ID: x}
                dual_vars = None
                is_mip = bool(inverse_data.get(self.IS_MIP, False))
                if y_kn is not None and not is_mip:
                    # Build dual vars dict keyed by constraint IDs
                    # Knitro returns duals for [eq_constrs; ineq_constrs]
                    y = np.array(y_kn)
                    n_eq = inverse_data[self.DIMS].zero
                    eq_dual = utilities.get_dual_values(
                        y[:n_eq],
                        utilities.extract_dual_value,
                        inverse_data[self.EQ_CONSTR])
                    ineq_dual = utilities.get_dual_values(
                        y[n_eq:],
                        utilities.extract_dual_value,
                        inverse_data[self.NEQ_CONSTR])
                    dual_vars = {}
                    dual_vars.update(eq_dual)
                    dual_vars.update(ineq_dual)
                solution = Solution(status, obj, primal_vars, dual_vars, attr)

        # Free the Knitro context.
        kn.KN_free(kc)
        return solution

    def solve_via_data(self, data, warm_start: bool, verbose: bool, solver_opts, solver_cache=None):
        """
        Returns the result of the call to the solver.

        Parameters
        ----------
        data : dict
            Data used by the solver.
        warm_start : bool
            Not used.
        verbose : bool
            Should the solver print output?
        solver_opts : dict
            Additional arguments for the solver.
        solver_cache: None
            None

        Returns
        -------
        The result of the call to the knitro solver.
        """
        import knitro as kn

        P = data[s.P]
        q = data[s.Q]
        A = data[s.A].tocoo()
        b = data[s.B]
        F = data[s.F].tocoo()
        g = data[s.G]
        lb = data[s.LOWER_BOUNDS]
        ub = data[s.UPPER_BOUNDS]

        results = {}
        try:
            kc = kn.KN_new()
        except Exception:  # Error in the Knitro.
            return {s.STATUS: s.SOLVER_ERROR}

        results[KNITRO.CONTEXT_KEY] = kc

        if not verbose:
            # Disable Knitro output.
            kn.KN_set_int_param(kc, kn.KN_PARAM_OUTLEV, kn.KN_OUTLEV_NONE)

        n_vars = 0
        if P is not None:
            n_vars = P.shape[0]
        elif q is not None:
            n_vars = q.shape[0]
        else:
            raise ValueError("No variables in the problem.")

        # Add n variables to the problem.
        kn.KN_add_vars(kc, n_vars)

        # Set the lower and upper bounds on the variables.
        if lb is not None:
            i, lb = kn_rm_inf(lb)
            kn.KN_set_var_lobnds(kc, indexVars=i, xLoBnds=lb)
        if ub is not None:
            i, ub = kn_rm_inf(ub)
            kn.KN_set_var_upbnds(kc, indexVars=i, xUpBnds=ub)

        # Set the variable types.
        # - default: KN_VARTYPE_CONTINUOUS.
        # - binray: KN_VARTYPE_BINARY.
        # - integer: KN_VARTYPE_INTEGER.
        vts = [kn.KN_VARTYPE_CONTINUOUS] * n_vars
        if s.BOOL_IDX in data:
            for j in data[s.BOOL_IDX]:
                vts[j] = kn.KN_VARTYPE_BINARY
        if s.INT_IDX in data:
            for j in data[s.INT_IDX]:
                vts[j] = kn.KN_VARTYPE_INTEGER
        kn.KN_set_var_types(kc, xTypes=vts)

        # Set the initial values of the primal variables.
        if KNITRO.X_INIT_KEY in solver_opts:
            i, vs = solver_opts[KNITRO.X_INIT_KEY]
            kn.KN_set_var_primal_init_values(kc, indexVars=i, xInitVals=vs)

        # Get the number of equality and inequality constraints.
        n_eqs, n_ineqs = A.shape[0], F.shape[0]
        n_cons = n_eqs + n_ineqs

        # Add the constraints to the problem.
        if n_cons > 0:
            kn.KN_add_cons(kc, n_cons)

        # Add linear equality and inequality constraints.
        if n_eqs > 0:
            cis = np.arange(n_eqs)
            kn.KN_set_con_eqbnds(kc, indexCons=cis, cEqBnds=b)
        if n_ineqs > 0:
            cis = n_eqs + np.arange(n_ineqs)
            kn.KN_set_con_upbnds(kc, indexCons=cis, cUpBnds=g)
        if n_eqs + n_ineqs > 0:
            D = sparse.vstack([A, F]).tocoo()
            cis, vis, coefs = D.row, D.col, D.data
            kn.KN_add_con_linear_struct(kc, indexCons=cis, indexVars=vis, coefs=coefs)

        # Set the initial values of the dual variables.
        if KNITRO.Y_INIT_KEY in solver_opts:
            i, vs = solver_opts[KNITRO.Y_INIT_KEY]
            kn.KN_set_con_dual_init_values(kc, indexCons=i, yInitVals=vs)

        # Set the objective function.
        # Set the linear part of the objective function.
        if q is not None:
            vis = np.arange(n_vars)
            kn.KN_add_obj_linear_struct(kc, indexVars=vis, coefs=q)

        # Set the quadratic part of the objective function.
        if P is not None and P.nnz != 0:
            Q = sparse.coo_matrix(0.5 * P)
            vis1, vis2, coefs = Q.row, Q.col, Q.data
            kn.KN_add_obj_quadratic_struct(kc, indexVars1=vis1, indexVars2=vis2, coefs=coefs)

        # Set the sense of the objective function.
        kn.KN_set_obj_goal(kc, kn.KN_OBJGOAL_MINIMIZE)

        # Set the values of the parameters.
        for k, v in solver_opts.items():
            if k in KNITRO.INTERFACE_ARGS:
                continue
            pid = kn.KN_get_param_id(kc, k)
            pt = kn.KN_get_param_type(kc, pid)
            fn = kn.KN_set_char_param
            if pt == kn.KN_PARAMTYPE_INTEGER:
                fn = kn.KN_set_int_param
            elif pt == kn.KN_PARAMTYPE_FLOAT:
                fn = kn.KN_set_double_param
            fn(kc, pid, v)

        # Optimize the problem.
        try:
            kn.KN_solve(kc)
        except Exception:  # Error in the solution
            results[s.STATUS] = s.SOLVER_ERROR

        # Cache the Knitro context.
        if solver_cache is not None:
            solver_cache[self.name()] = kc

        return results

    def cite(self, data):
        """Returns bibtex citation for the solver.

        Parameters
        ----------
        data : dict
            Data generated via an apply call.
        """
        return CITATION_DICT["KNITRO"]



"""
Copyright 2013 Steven Diamond

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import cvxpy.lin_ops.lin_utils as lu
import cvxpy.utilities as u
from cvxpy.error import DCPError
from cvxpy.expressions.expression import Expression
from cvxpy.interface.matrix_utilities import scalar_value
from cvxpy.utilities import scopes


class Objective(u.Canonical):
    """An optimization objective.

    Parameters
    ----------
    expr : Expression
        The expression to act upon. Must be a scalar.

    Raises
    ------
    ValueError
        If expr is not a scalar.
    """

    NAME = "objective"

    def __init__(self, expr) -> None:
        self.args = [Expression.cast_to_const(expr)]
        # Needed for Canonical._aggregate_metrics.
        self.ndim = 0
        # Validate that the objective resolves to a scalar.
        if not self.args[0].is_scalar():
            raise ValueError(f"The '{self.NAME}' objective must resolve to a scalar.")
        if not self.args[0].is_real():
            raise ValueError(f"The '{self.NAME}' objective must be real valued.")

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({repr(self.args[0])})"

    def __str__(self) -> str:
        return f'{self.NAME} {self.args[0].name()}'
    
    def format_labeled(self):
        """Format objective with labels where available."""
        return f'{self.NAME} {self.args[0].format_labeled()}'

    def __radd__(self, other):
        if other == 0:
            return self
        else:
            raise NotImplementedError()

    def __sub__(self, other):
        if not isinstance(other, (Minimize, Maximize)):
            raise NotImplementedError()
        # Objectives must opposites
        return self + (-other)

    def __rsub__(self, other):
        if other == 0:
            return -self
        else:
            raise NotImplementedError()

    def __mul__(self, other):
        if not isinstance(other, (int, float)):
            raise NotImplementedError()
        # If negative, reverse the direction of objective
        if (type(self) == Maximize) == (other < 0.0):
            return Minimize(self.args[0] * other)
        else:
            return Maximize(self.args[0] * other)

    __rmul__ = __mul__

    def __div__(self, other):
        if not isinstance(other, (int, float)):
            raise NotImplementedError()
        return self * (1.0/other)

    __truediv__ = __div__

    @property
    def value(self):
        """The value of the objective expression.
        """
        v = self.args[0].value
        if v is None:
            return None
        else:
            return scalar_value(v)

    def is_quadratic(self) -> bool:
        """Returns if the objective is a quadratic function.
        """
        return self.args[0].is_quadratic()

    def is_qpwa(self) -> bool:
        """Returns if the objective is a quadratic of piecewise affine.
        """
        return self.args[0].is_qpwa()


class Minimize(Objective):
    """An optimization objective for minimization.

    Parameters
    ----------
    expr : Expression
        The expression to minimize. Must be a scalar.

    Raises
    ------
    ValueError
        If expr is not a scalar.
    """

    NAME = "minimize"

    def __neg__(self) -> "Maximize":
        return Maximize(-self.args[0])

    def __add__(self, other):
        if not isinstance(other, (Minimize, Maximize)):
            raise NotImplementedError()
        # Objectives must both be Minimize.
        if type(other) is Minimize:
            return Minimize(self.args[0] + other.args[0])
        else:
            raise DCPError("Problem does not follow DCP rules.")

    def canonicalize(self):
        """Pass on the target expression's objective and constraints.
        """
        return self.args[0].canonical_form

    def is_dcp(self, dpp: bool = False) -> bool:
        """The objective must be convex.
        """
        if dpp:
            with scopes.dpp_scope():
                return self.args[0].is_convex()
        return self.args[0].is_convex()

    def is_dgp(self, dpp: bool = False) -> bool:
        """The objective must be log-log convex.
        """
        if dpp:
            with scopes.dpp_scope():
                return self.args[0].is_log_log_convex()
        return self.args[0].is_log_log_convex()

    def is_dpp(self, context='dcp') -> bool:
        with scopes.dpp_scope():
            if context.lower() == 'dcp':
                return self.is_dcp(dpp=True)
            elif context.lower() == 'dgp':
                return self.is_dgp(dpp=True)
            else:
                raise ValueError("Unsupported context ", context)

    def is_dqcp(self) -> bool:
        """The objective must be quasiconvex.
        """
        return self.args[0].is_quasiconvex()

    @staticmethod
    def primal_to_result(result):
        """The value of the objective given the solver primal value.
        """
        return result


class Maximize(Objective):
    """An optimization objective for maximization.

    Parameters
    ----------
    expr : Expression
        The expression to maximize. Must be a scalar.

    Raises
    ------
    ValueError
        If expr is not a scalar.
    """

    NAME = "maximize"

    def __neg__(self) -> Minimize:
        return Minimize(-self.args[0])

    def __add__(self, other):
        if not isinstance(other, (Minimize, Maximize)):
            raise NotImplementedError()
        # Objectives must both be Maximize.
        if type(other) is Maximize:
            return Maximize(self.args[0] + other.args[0])
        else:
            raise DCPError("Problem does not follow DCP rules.")

    def canonicalize(self):
        """Negates the target expression's objective.
        """
        obj, constraints = self.args[0].canonical_form
        return (lu.neg_expr(obj), constraints)

    def is_dcp(self, dpp: bool = False) -> bool:
        """The objective must be concave.
        """
        if dpp:
            with scopes.dpp_scope():
                return self.args[0].is_concave()
        return self.args[0].is_concave()

    def is_dgp(self, dpp: bool = False) -> bool:
        """The objective must be log-log concave.
        """
        if dpp:
            with scopes.dpp_scope():
                return self.args[0].is_log_log_concave()
        return self.args[0].is_log_log_concave()

    def is_dpp(self, context='dcp') -> bool:
        with scopes.dpp_scope():
            if context.lower() == 'dcp':
                return self.is_dcp(dpp=True)
            elif context.lower() == 'dgp':
                return self.is_dgp(dpp=True)
            else:
                raise ValueError("Unsupported context ", context)

    def is_dqcp(self) -> bool:
        """The objective must be quasiconcave.
        """
        return self.args[0].is_quasiconcave()

    @staticmethod
    def primal_to_result(result):
        """The value of the objective given the solver primal value.
        """
        return -result



"""
Copyright 2025 CVXPY developers

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
import scipy.sparse as sp

import cvxpy.settings as s
from cvxpy.constraints import PSD, SOC, ExpCone, PowCone3D
from cvxpy.reductions.dcp2cone.cone_matrix_stuffing import ParamConeProg
from cvxpy.reductions.solution import Solution, failure_solution
from cvxpy.reductions.solvers import utilities
from cvxpy.reductions.solvers.conic_solvers.conic_solver import ConicSolver, dims_to_solver_dict
from cvxpy.utilities.citations import CITATION_DICT


def kn_isinf(x) -> bool:
    """Check if x is -inf or inf."""
    if x <= -np.inf or x >= np.inf:
        return True
    if x <= float("-inf") or x >= float("inf"):
        return True

    import knitro as kn

    if x <= -kn.KN_INFINITY or x >= kn.KN_INFINITY:
        return True
    return False


def kn_rm_inf(a) -> tuple[list[int], list[float]]:
    """Convert -inf to -kn.KN_INFINITY and inf to kn.KN_INFINITY."""
    i, vs = [], []
    for j, v in enumerate(a):
        if not kn_isinf(v):
            i.append(j)
            vs.append(v)
    return i, vs


class Dims:
    def __init__(self, dims: dict):
        self.n_eq_cons = int(dims.get(s.EQ_DIM, 0))
        self.n_ineq_cons = int(dims.get(s.LEQ_DIM, 0))
        self.soc_dims = [int(d) for d in dims.get(s.SOC_DIM, [])]
        self.n_exps = int(dims.get(s.EXP_DIM, 0))
        self.p3d_exps = dims.get("p", [])
        self.psd_dims = dims.get(s.PSD_DIM, [])
        self.n_p3ds = len(self.p3d_exps)
        self.n_socs = len(self.soc_dims)
        self.n_psds = len(self.psd_dims)
        self.n_exp_cons = self.n_exps
        self.n_p3d_cons = self.n_p3ds
        self.n_soc_cons = self.n_socs
        self.n_psd_cons = sum(d**2 for d in self.psd_dims)
        self.n_cone_cons = self.n_soc_cons + self.n_exp_cons + self.n_psd_cons + self.n_p3d_cons
        self.n_soc_vars = sum(self.soc_dims)
        self.n_exp_vars = 3 * self.n_exp_cons
        self.n_psd_vars = sum(d**2 for d in self.psd_dims)
        self.n_p3d_vars = 3 * self.n_p3d_cons
        self.n_cone_vars = self.n_soc_vars + self.n_exp_vars + self.n_psd_vars + self.n_p3d_vars


class CB:
    # Knitro callback
    def __init__(self, f, grad=None, hess=None):
        self.f = f
        self.grad = grad
        self.hess = hess


class Ctx:
    def __init__(self, n: int, vp: int, cp: int, a=None):
        self.n = n  # Number of constraints
        self.vp = vp  # Variable position
        self.cp = cp  # Constraint position
        if a is not None:
            self.a = np.array(a)


def build_exp_cb() -> CB:
    import knitro as kn

    def f(_, cb: kn.CB_context, req: kn.KN_eval_request, res: kn.KN_eval_result, ctx: Ctx):
        if req.type != kn.KN_RC_EVALFC:
            return -1
        n = ctx.n
        vp = ctx.vp
        v = np.asarray(req.x[vp : vp + 3 * n])
        x = v[0::3]
        y = v[1::3]
        z = v[2::3]
        xy = np.divide(x, y)
        with np.errstate(over="ignore"):
            exp = np.exp(xy)
        mask = (np.isclose(y, 0.0)) | np.isnan(xy) | np.isnan(exp) | np.isinf(exp)
        ind = np.where(x <= 0.0, -1.0, 0.0)
        res.c[:n] = np.where(mask, ind, y * exp - z)
        return 0

    def grad(_, cb: kn.CB_context, req: kn.KN_eval_request, res: kn.KN_eval_result, ctx: Ctx):
        if req.type != kn.KN_RC_EVALGA:
            return -1
        n = ctx.n
        vp = ctx.vp
        v = np.asarray(req.x[vp : vp + 3 * n])
        x = v[0::3]
        y = v[1::3]
        xy = np.divide(x, y)
        with np.errstate(over="ignore"):
            exp = np.exp(xy)
        mask = (np.isclose(y, 0.0)) | np.isnan(xy) | np.isnan(exp) | np.isinf(exp)
        res.jac[0::3] = np.where(mask, np.inf, exp)
        res.jac[1::3] = np.where(mask, np.inf, (1 - xy) * exp)
        res.jac[2::3] = np.where(mask, np.inf, -1.0)
        return 0

    def hess(_, cb: kn.CB_context, req: kn.KN_eval_request, res: kn.KN_eval_result, ctx: Ctx):
        if req.type != kn.KN_RC_EVALH and req.type != kn.KN_RC_EVALH_NO_F:
            return -1
        n = ctx.n
        vp = ctx.vp
        cp = ctx.cp
        v = np.asarray(req.x[vp : vp + 3 * n])
        u = np.asarray(req.lambda_[cp : cp + n])
        x = v[0::3]
        y = v[1::3]
        xy = np.divide(x, y)
        with np.errstate(over="ignore"):
            exp = np.exp(xy)
        mask = (np.isclose(y, 0.0)) | np.isnan(xy) | np.isnan(exp) | np.isinf(exp)
        iy = np.divide(1, y)
        hx = iy * exp * u
        res.hess[0::3] = np.where(mask, np.inf, hx)
        res.hess[1::3] = np.where(mask, np.inf, -xy * hx)
        res.hess[2::3] = np.where(mask, np.inf, (xy**2) * hx)
        return 0

    return CB(f=f, grad=grad, hess=hess)


def build_pow3d_cb() -> CB:
    import knitro as kn

    def f(_, cb: kn.CB_context, req: kn.KN_eval_request, res: kn.KN_eval_result, ctx: Ctx):
        if req.type != kn.KN_RC_EVALFC:
            return -1
        n = ctx.n
        vp = ctx.vp
        a = ctx.a
        v = np.asarray(req.x[vp : vp + 3 * n])
        x = v[0::3]
        y = v[1::3]
        z = v[2::3]
        res.c[:n] = -np.power(x, a) * np.power(y, 1 - a) + np.abs(z)
        return 0

    def grad(_, cb: kn.CB_context, req: kn.KN_eval_request, res: kn.KN_eval_result, ctx: Ctx):
        if req.type != kn.KN_RC_EVALGA:
            return -1
        n = ctx.n
        vp = ctx.vp
        a = ctx.a
        v = np.asarray(req.x[vp : vp + 3 * n])
        x = v[0::3]
        y = v[1::3]
        z = v[2::3]
        res.jac[0::3] = -a * np.power(x, a - 1) * np.power(y, 1 - a)
        res.jac[1::3] = -(1 - a) * np.power(x, a) * np.power(y, -a)
        res.jac[2::3] = np.sign(z)
        return 0

    def hess(_, cb: kn.CB_context, req: kn.KN_eval_request, res: kn.KN_eval_result, ctx: Ctx):
        if req.type != kn.KN_RC_EVALH and req.type != kn.KN_RC_EVALH_NO_F:
            return -1
        n = ctx.n
        vp = ctx.vp
        cp = ctx.cp
        a = ctx.a
        b = a * (1 - a)
        v = np.asarray(req.x[vp : vp + 3 * n])
        u = np.asarray(req.lambda_[cp : cp + n])
        x = v[0::3]
        y = v[1::3]
        res.hess[0::3] = b * np.power(x, a - 2) * np.power(y, 1 - a) * u
        res.hess[1::3] = -b * np.power(x, a - 1) * np.power(y, -a) * u
        res.hess[2::3] = b * np.power(x, a) * np.power(y, -a - 1) * u
        return 0

    return CB(f=f, grad=grad, hess=hess)


class KNITRO(ConicSolver):
    """
    Conic interface for the Knitro solver.
    """

    # Solver capabilities.
    MIP_CAPABLE = True
    BOUNDED_VARIABLES = True
    SUPPORTED_CONSTRAINTS = ConicSolver.SUPPORTED_CONSTRAINTS + [SOC, ExpCone, PowCone3D, PSD]
    MI_SUPPORTED_CONSTRAINTS = SUPPORTED_CONSTRAINTS

    # Keys:
    CONTEXT_KEY = "context"
    X_INIT_KEY = "x_init"
    Y_INIT_KEY = "y_init"
    N_VARS_KEY = "n_vars"
    N_CONS_KEY = "n_cons"

    # Keyword arguments for the CVXPY interface.
    INTERFACE_ARGS = [X_INIT_KEY, Y_INIT_KEY]

    EXP_CONE_ORDER = [0, 1, 2]
    EXP_DOUBLE_LIMIT = 705.0

    # Map of Knitro status to CVXPY status.
    # This is based on the Knitro documentation:
    # https://www.artelys.com/app/docs/knitro/3_referenceManual/returnCodes.html
    STATUS_MAP = {
        0: s.OPTIMAL,
        -100: s.OPTIMAL_INACCURATE,
        -101: s.USER_LIMIT,
        -102: s.USER_LIMIT,
        -103: s.USER_LIMIT,
        -200: s.INFEASIBLE,
        -201: s.INFEASIBLE,
        -202: s.INFEASIBLE,
        -203: s.INFEASIBLE,
        -204: s.INFEASIBLE,
        -205: s.INFEASIBLE,
        -300: s.UNBOUNDED,
        -301: s.UNBOUNDED,
        -400: s.USER_LIMIT,
        -401: s.USER_LIMIT,
        -402: s.USER_LIMIT,
        -403: s.USER_LIMIT,
        -404: s.USER_LIMIT,
        -405: s.USER_LIMIT,
        -406: s.USER_LIMIT,
        -410: s.USER_LIMIT,
        -411: s.USER_LIMIT,
        -412: s.USER_LIMIT,
        -413: s.USER_LIMIT,
        -415: s.USER_LIMIT,
        -416: s.USER_LIMIT,
        -500: s.SOLVER_ERROR,
        -501: s.SOLVER_ERROR,
        -502: s.SOLVER_ERROR,
        -503: s.SOLVER_ERROR,
        -504: s.SOLVER_ERROR,
        -505: s.SOLVER_ERROR,
        -506: s.SOLVER_ERROR,
        -507: s.SOLVER_ERROR,
        -508: s.SOLVER_ERROR,
        -509: s.SOLVER_ERROR,
        -510: s.SOLVER_ERROR,
        -511: s.SOLVER_ERROR,
        -512: s.SOLVER_ERROR,
        -513: s.SOLVER_ERROR,
        -514: s.SOLVER_ERROR,
        -515: s.SOLVER_ERROR,
        -516: s.SOLVER_ERROR,
        -517: s.SOLVER_ERROR,
        -518: s.SOLVER_ERROR,
        -519: s.SOLVER_ERROR,
        -520: s.SOLVER_ERROR,
        -521: s.SOLVER_ERROR,
        -522: s.SOLVER_ERROR,
        -523: s.SOLVER_ERROR,
        -524: s.SOLVER_ERROR,
        -525: s.SOLVER_ERROR,
        -526: s.SOLVER_ERROR,
        -527: s.SOLVER_ERROR,
        -528: s.SOLVER_ERROR,
        -529: s.SOLVER_ERROR,
        -530: s.SOLVER_ERROR,
        -531: s.SOLVER_ERROR,
        -532: s.SOLVER_ERROR,
        -600: s.SOLVER_ERROR,
    }

    def name(self):
        """The name of the solver."""
        return s.KNITRO

    def import_solver(self) -> None:
        """Imports the solver."""
        import knitro  # noqa: F401

    def supports_quad_obj(self):
        return True

    def accepts(self, problem) -> bool:
        return super(KNITRO, self).accepts(problem)

    def apply(self, problem: ParamConeProg):
        """Returns a new problem and data for inverting the new solution.

        Returns
        -------
        tuple
            (dict of arguments needed for the solver, inverse data)
        """
        data, inv_data = super(KNITRO, self).apply(problem)
        variables = problem.x
        data[s.BOOL_IDX] = [int(t[0]) for t in variables.boolean_idx]
        data[s.INT_IDX] = [int(t[0]) for t in variables.integer_idx]
        inv_data["is_mip"] = data[s.BOOL_IDX] or data[s.INT_IDX]
        return data, inv_data

    def invert(self, results, inverse_data):
        """Returns the solution to the original problem given the inverse_data."""
        import knitro as kn

        if KNITRO.CONTEXT_KEY not in results:
            return failure_solution(s.SOLVER_ERROR)

        kc = results[KNITRO.CONTEXT_KEY]
        num_iters = kn.KN_get_number_iters(kc)
        solve_time = kn.KN_get_solve_time_real(kc)
        attr = {
            s.SOLVE_TIME: solve_time,
            s.NUM_ITERS: num_iters,
            s.EXTRA_STATS: kc,
        }

        if s.STATUS in results and results[s.STATUS] == s.SOLVER_ERROR:
            solution = failure_solution(s.SOLVER_ERROR, attr)
        else:
            status_kn, obj_kn, x_kn, y_kn = kn.KN_get_solution(kc)
            status = self.STATUS_MAP.get(status_kn, s.SOLVER_ERROR)

            if status == s.UNBOUNDED:
                solution = Solution(status, -np.inf, {}, {}, attr)
            elif (status not in s.SOLUTION_PRESENT) or (x_kn is None):
                solution = failure_solution(status, attr)
            else:
                n_vars = int(results[KNITRO.N_VARS_KEY])
                n_cons = int(results[KNITRO.N_CONS_KEY])

                obj = obj_kn + inverse_data[s.OFFSET]
                x_kn = x_kn[:n_vars]
                x = np.array(x_kn)
                primal_vars = {inverse_data[KNITRO.VAR_ID]: x}

                dual_vars = None
                is_mip = bool(inverse_data.get("is_mip", False))
                y_kn = kn.KN_get_con_dual_values(kc)
                if y_kn is not None and not is_mip:
                    dims = dims_to_solver_dict(inverse_data[s.DIMS] or {})
                    y_kn = y_kn[:n_cons]
                    n_eqs = int(dims.get(s.EQ_DIM, 0))
                    y = np.array(y_kn)
                    eq_dual_vars = utilities.get_dual_values(
                        y[:n_eqs],
                        utilities.extract_dual_value,
                        inverse_data[KNITRO.EQ_CONSTR],
                    )
                    ineq_dual_vars = utilities.get_dual_values(
                        y[n_eqs:],
                        utilities.extract_dual_value,
                        inverse_data[KNITRO.NEQ_CONSTR],
                    )
                    dual_vars = {**eq_dual_vars, **ineq_dual_vars}
                solution = Solution(status, obj, primal_vars, dual_vars, attr)
        # Free the Knitro context.
        kn.KN_free(kc)
        return solution

    def solve_via_data(self, data, warm_start: bool, verbose: bool, solver_opts, solver_cache=None):
        """Returns the result of the call to the solver.

        Parameters
        ----------
        data : dict
            Data used by the solver.
        warm_start : bool
            Not used.
        verbose : bool
            Should the solver print output?
        solver_opts : dict
            Additional arguments for the solver.

        Returns
        -------
        The result of the call to the knitro solver.
        """
        import knitro as kn

        P = data.get(s.P)
        c = data.get(s.C)
        b = data.get(s.B)
        A = data.get(s.A)
        lb = data.get(s.LOWER_BOUNDS)
        ub = data.get(s.UPPER_BOUNDS)
        dims = Dims(dims_to_solver_dict(data.get(s.DIMS)))

        results = {}
        try:
            kc = kn.KN_new()
        except Exception:
            results[s.STATUS] = s.SOLVER_ERROR
            return results

        results[KNITRO.CONTEXT_KEY] = kc

        if not verbose:
            # Disable Knitro output.
            kn.KN_set_int_param(kc, kn.KN_PARAM_OUTLEV, kn.KN_OUTLEV_NONE)

        n_vars = int(c.shape[0])
        results[KNITRO.N_VARS_KEY] = n_vars

        # Add n variables to the problem.
        kn.KN_add_vars(kc, n_vars)

        # Set the lower and upper bounds on the variables.
        if lb is not None:
            i, lb = kn_rm_inf(lb)
            kn.KN_set_var_lobnds(kc, indexVars=i, xLoBnds=lb)
        if ub is not None:
            i, ub = kn_rm_inf(ub)
            kn.KN_set_var_upbnds(kc, indexVars=i, xUpBnds=ub)

        # Set the variable types.
        vts = [kn.KN_VARTYPE_CONTINUOUS] * n_vars
        if s.BOOL_IDX in data:
            for j in data[s.BOOL_IDX]:
                vts[j] = kn.KN_VARTYPE_BINARY
        if s.INT_IDX in data:
            for j in data[s.INT_IDX]:
                vts[j] = kn.KN_VARTYPE_INTEGER
        kn.KN_set_var_types(kc, xTypes=vts)

        # Set the initial values of the primal variables.
        if KNITRO.X_INIT_KEY in solver_opts:
            i, vs = solver_opts[KNITRO.X_INIT_KEY]
            kn.KN_set_var_primal_init_values(kc, indexVars=i, xInitVals=vs)

        # Add constraints to the problem.
        n_cons = int(A.shape[0]) if A is not None else 0
        results[KNITRO.N_CONS_KEY] = n_cons
        if n_cons > 0:
            kn.KN_add_cons(kc, n_cons)

        if dims.n_cone_vars > 0:
            kn.KN_add_vars(kc, dims.n_cone_vars)
        if dims.n_cone_cons > 0:
            kn.KN_add_cons(kc, dims.n_cone_cons)
        if dims.n_psds > 0:
            kn.KN_add_vars(kc, dims.n_psd_vars)

        D = sp.coo_matrix(A)
        if D.nnz != 0:
            cis, vis, coefs = D.row, D.col, D.data
            kn.KN_add_con_linear_struct(kc, indexCons=cis, indexVars=vis, coefs=coefs)

        vp, cp = 0, 0
        if dims.n_eq_cons > 0:
            cis = cp + np.arange(dims.n_eq_cons)
            kn.KN_set_con_eqbnds(kc, indexCons=cis, cEqBnds=b[cis])
        cp += dims.n_eq_cons

        if dims.n_ineq_cons > 0:
            cis = cp + np.arange(dims.n_ineq_cons)
            kn.KN_set_con_upbnds(kc, indexCons=cis, cUpBnds=b[cis])
        cp += dims.n_ineq_cons
        vp += n_vars

        if dims.n_cone_vars > 0:
            vis = vp + np.arange(dims.n_cone_vars)
            cis = cp + np.arange(dims.n_cone_vars)
            coefs = np.ones_like(vis)
            kn.KN_set_con_eqbnds(kc, indexCons=cis, cEqBnds=b[cis])
            kn.KN_add_con_linear_struct(kc, indexCons=cis, indexVars=vis, coefs=coefs)
        cp += dims.n_cone_vars

        if dims.n_socs > 0:
            cis = cp + np.arange(dims.n_soc_cons)
            vis = vp + np.insert(np.cumsum(dims.soc_dims), 0, 0)[:-1]
            bnds = np.zeros_like(cis)
            coefs = -np.ones_like(vis)
            kn.KN_set_con_upbnds(kc, indexCons=cis, cUpBnds=bnds)
            kn.KN_set_var_lobnds(kc, indexVars=vis, xLoBnds=bnds)
            kn.KN_add_con_quadratic_struct(
                kc, indexCons=cis, indexVars1=vis, indexVars2=vis, coefs=coefs
            )

        for k in range(dims.n_soc_cons):
            d = dims.soc_dims[k]
            vis = vp + np.arange(1, d)
            coefs = np.ones_like(vis)
            kn.KN_add_con_quadratic_struct(
                kc, indexCons=cp + k, indexVars1=vis, indexVars2=vis, coefs=coefs
            )
            vp += d
        cp += dims.n_soc_cons

        if dims.n_exps > 0:
            cis = cp + np.arange(dims.n_exp_cons)
            vis = vp + np.arange(dims.n_exp_vars)
            bnds = np.zeros_like(cis)
            kn.KN_set_con_upbnds(kc, indexCons=cis, cUpBnds=bnds)
            kn.KN_set_var_lobnds(kc, indexVars=vis[1::3], xLoBnds=bnds)
            kn.KN_set_var_lobnds(kc, indexVars=vis[2::3], xLoBnds=bnds)

            cb = build_exp_cb()

            kb = kn.KN_add_eval_callback(kc, indexCons=cis, funcCallback=cb.f)
            jcis = np.repeat(cis, 3)
            jvis = vis
            kn.KN_set_cb_grad(kc, kb, jacIndexCons=jcis, jacIndexVars=jvis, gradCallback=cb.grad)
            hvis = np.repeat(vis[0::3], 3)
            hvis1 = hvis + np.tile(np.array([0, 0, 1]), dims.n_exps)
            hvis2 = hvis + np.tile(np.array([0, 1, 1]), dims.n_exps)
            kn.KN_set_cb_hess(
                kc, kb, hessIndexVars1=hvis1, hessIndexVars2=hvis2, hessCallback=cb.hess
            )
            ctx = Ctx(n=dims.n_exps, vp=vp, cp=cp)
            kn.KN_set_cb_user_params(kc, kb, ctx)
        cp += dims.n_exp_cons
        vp += dims.n_exp_vars

        if dims.n_psds > 0:
            cis = cp + np.arange(dims.n_psd_cons)
            vis = vp + np.arange(dims.n_psd_vars)
            bnds = np.zeros_like(cis)
            coefs = -np.ones_like(vis)
            kn.KN_set_con_eqbnds(kc, indexCons=cis, cEqBnds=bnds)
            kn.KN_add_con_linear_struct(kc, indexCons=cis, indexVars=vis, coefs=coefs)
            vp += dims.n_psd_vars

        vp += dims.n_p3d_vars
        for k in range(dims.n_psds):
            d = dims.psd_dims[k]
            vis = vp + np.arange(d**2)
            cis = cp + np.arange(d**2)
            for i in range(d):
                for j in range(d):
                    vis1 = vis[d * i : d * (i + 1)]
                    vis2 = vis[d * j : d * (j + 1)]
                    coefs = np.ones_like(vis1)
                    kn.KN_add_con_quadratic_struct(
                        kc, indexCons=cis[d * i + j], indexVars1=vis1, indexVars2=vis2, coefs=coefs
                    )
            cp += d**2
            vp += d**2
        vp -= dims.n_p3d_vars

        if dims.n_p3ds > 0:
            cis = cp + np.arange(dims.n_p3d_cons)
            vis = vp + np.arange(dims.n_p3d_vars)
            bnds = np.zeros_like(cis)
            kn.KN_set_con_upbnds(kc, indexCons=cis, cUpBnds=bnds)
            kn.KN_set_var_lobnds(kc, indexVars=vis[0::3], xLoBnds=bnds)
            kn.KN_set_var_lobnds(kc, indexVars=vis[1::3], xLoBnds=bnds)

            cb = build_pow3d_cb()

            kb = kn.KN_add_eval_callback(kc, indexCons=cis, funcCallback=cb.f)
            jcis = np.repeat(cis, 3)
            jvis = vis
            kn.KN_set_cb_grad(kc, kb, jacIndexCons=jcis, jacIndexVars=jvis, gradCallback=cb.grad)
            hvis = np.repeat(vis[0::3], 3)
            hvis1 = hvis + np.tile(np.array([0, 0, 1]), dims.n_p3ds)
            hvis2 = hvis + np.tile(np.array([0, 1, 1]), dims.n_p3ds)
            kn.KN_set_cb_hess(
                kc, kb, hessIndexVars1=hvis1, hessIndexVars2=hvis2, hessCallback=cb.hess
            )
            ctx = Ctx(n=dims.n_p3ds, vp=vp, cp=cp, a=dims.p3d_exps)
            kn.KN_set_cb_user_params(kc, kb, ctx)
        cp += dims.n_p3d_cons
        vp += dims.n_p3d_vars

        # Set the initial values of the dual variables.
        if KNITRO.Y_INIT_KEY in solver_opts:
            i, vs = solver_opts[KNITRO.Y_INIT_KEY]
            kn.KN_set_con_dual_init_values(kc, indexCons=i, yInitVals=vs)

        # Set the linear part of the objective function.
        if c is not None:
            vis = np.arange(n_vars)
            kn.KN_add_obj_linear_struct(kc, indexVars=vis, coefs=c)

        # Set the quadratic part of the objective function.
        if P is not None and P.nnz != 0:
            Q = sp.coo_matrix(0.5 * P)
            vis1, vis2, coefs = Q.row, Q.col, Q.data
            kn.KN_add_obj_quadratic_struct(kc, indexVars1=vis1, indexVars2=vis2, coefs=coefs)

        # Set the sense of the objective function.
        kn.KN_set_obj_goal(kc, kn.KN_OBJGOAL_MINIMIZE)

        # Set the values of the parameters.
        for k, v in solver_opts.items():
            if k in KNITRO.INTERFACE_ARGS:
                continue
            pid = kn.KN_get_param_id(kc, k)
            pt = kn.KN_get_param_type(kc, pid)
            fn = kn.KN_set_char_param
            if pt == kn.KN_PARAMTYPE_INTEGER:
                fn = kn.KN_set_int_param
            elif pt == kn.KN_PARAMTYPE_FLOAT:
                fn = kn.KN_set_double_param
            fn(kc, pid, v)

        # Optimize the problem.
        try:
            kn.KN_solve(kc)
        except Exception:  # Error in the solution
            results[s.STATUS] = s.SOLVER_ERROR

        # Cache the Knitro context.
        if solver_cache is not None:
            solver_cache[self.name()] = kc

        return results

    def cite(self, data):
        """Returns bibtex citation for the solver.

        Parameters
        ----------
        data : dict
            Data generated via an apply call.
        """
        return CITATION_DICT[s.KNITRO]



"""
Copyright 2017 Robin Verschueren

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
from numpy.lib.array_utils import normalize_axis_index, normalize_axis_tuple
from scipy.sparse import eye_array

from cvxpy.atoms.quad_form import SymbolicQuadForm
from cvxpy.expressions.variable import Variable


def quad_over_lin_canon(expr, args, solver_context=None):
    affine_expr = args[0]
    y = args[1]
    assert y.is_scalar(), "quad_over_lin requires scalar y"
    axis = expr.axis

    # Simplify if y has no parameters.
    if len(y.parameters()) == 0:
        quad_mat = eye_array(affine_expr.size) / y.value
    else:
        # TODO this codepath produces an intermediate dense matrix.
        # but it should be sparse the whole time.
        quad_mat = eye_array(affine_expr.size) / y

    # Compute block_indices if axis is specified
    block_indices = None
    if axis is not None:
        shape = affine_expr.shape
        ndim = len(shape)

        # Normalize axis/axes to tuple
        if isinstance(axis, int):
            axes = (normalize_axis_index(axis, ndim),)
        else:
            axes = normalize_axis_tuple(axis, ndim)

        block_indices = _compute_block_indices(shape, axes)

    if isinstance(affine_expr, Variable):
        return SymbolicQuadForm(affine_expr, quad_mat, expr, block_indices=block_indices), []
    else:
        t = Variable(affine_expr.shape)
        return SymbolicQuadForm(t, quad_mat, expr, block_indices=block_indices), [affine_expr == t]


def _compute_block_indices(shape, axes):
    """Compute block indices for reducing along specified axes (Fortran order).

    Parameters
    ----------
    shape : tuple
        Shape of input array
    axes : tuple of int
        Axes being reduced (must be normalized/positive)

    Returns
    -------
    list of np.ndarray
        block_indices[j] = array of input indices for output element j
    """
    ndim = len(shape)
    axes_set = set(axes)

    # Output shape (dimensions not in axes)
    output_dims = [i for i in range(ndim) if i not in axes_set]
    output_shape = tuple(shape[i] for i in output_dims)
    n_outputs = int(np.prod(output_shape)) if output_shape else 1

    # Dimensions being reduced
    reduce_shape = tuple(shape[i] for i in sorted(axes))
    reduce_size = int(np.prod(reduce_shape))

    # Vectorized computation of all indices at once
    # out_multi: tuple of arrays, each shape (n_outputs,)
    if output_shape:
        out_multi = np.unravel_index(np.arange(n_outputs), output_shape, order="F")
    else:
        out_multi = ()

    # reduce_multi: tuple of arrays, each shape (reduce_size,)
    if reduce_shape:
        reduce_multi = np.unravel_index(np.arange(reduce_size), reduce_shape, order="F")
    else:
        reduce_multi = ()

    # Build input multi-index arrays with broadcasting
    # Each element will have shape (n_outputs, reduce_size)
    input_multi = []
    out_ptr = 0
    reduce_ptr = 0
    for i in range(ndim):
        if i in axes_set:
            # Broadcast reduce indices: (reduce_size,) -> (n_outputs, reduce_size)
            input_multi.append(np.broadcast_to(reduce_multi[reduce_ptr], (n_outputs, reduce_size)))
            reduce_ptr += 1
        else:
            # Broadcast output indices: (n_outputs,) -> (n_outputs, reduce_size)
            input_multi.append(
                np.broadcast_to(out_multi[out_ptr][:, np.newaxis], (n_outputs, reduce_size))
            )
            out_ptr += 1

    # Compute all flat indices at once: shape (n_outputs, reduce_size)
    flat_indices_2d = np.ravel_multi_index(input_multi, shape, order="F")

    # Return as list of arrays for compatibility with SymbolicQuadForm
    return [flat_indices_2d[j] for j in range(n_outputs)]



"""
Copyright, the CVXPY authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import math

from cvxpy.atoms.affine.hstack import hstack
from cvxpy.atoms.affine.promote import promote
from cvxpy.atoms.affine.reshape import reshape
from cvxpy.atoms.affine.transpose import transpose
from cvxpy.atoms.affine.vec import vec
from cvxpy.atoms.affine.vstack import vstack
from cvxpy.atoms.log_sum_exp import log_sum_exp


def pnorm_canon(expr, args):
    x = args[0]
    p = expr.original_p
    if x.shape == tuple():
        x = promote(x, (1,))
    if expr.axis is None or len(x.shape) == 1:
        x = vec(x, order='F')
        return (1.0/p) * log_sum_exp(hstack([xi * p for xi in x])), []

    axis = expr.axis
    axes = (axis,) if isinstance(axis, int) else tuple(axis)
    ndim = len(x.shape)

    # Permute so non-reduce axes come first, reduce axes come last
    keep = [i for i in range(ndim) if i not in axes]
    perm = keep + list(axes)
    x_perm = transpose(x, axes=perm) if perm != list(range(ndim)) else x

    # Reshape to 2D: (n_output, n_reduce)
    n_output = math.prod(x.shape[i] for i in keep)
    n_reduce = math.prod(x.shape[i] for i in axes)
    x_2d = reshape(x_perm, (n_output, n_reduce), order='F')

    rows = []
    for i in range(n_output):
        row = x_2d[i]
        rows.append((1.0/p) * log_sum_exp(hstack([xi * p for xi in row])))
    return reshape(vstack(rows), expr.shape, order='F'), []



from cvxpy.expressions.variable import Variable
from cvxpy.utilities import scopes
from cvxpy.utilities.solver_context import SolverInfo


# TODO(akshayka): expose as a reduction for user's convenience
def mul_canon(expr, args, solver_context: SolverInfo | None = None):
    # Only allow param * var (not var * param). Associate right to left.
    # TODO: Only descend if both sides have parameters
    lhs = args[0]
    rhs = args[1]
    if not (lhs.parameters() and rhs.parameters()):
        return expr.copy(args), []

    op_type = type(expr)
    if lhs.variables():
        with scopes.dpp_scope():
            assert rhs.is_affine()
        t = Variable(lhs.shape)
        return op_type(t, rhs), [t == lhs]
    elif rhs.variables():
        with scopes.dpp_scope():
            assert lhs.is_affine()
        t = Variable(rhs.shape)
        return op_type(lhs, t), [t == rhs]

    # Neither side has variables. One side must be affine in parameters.
    lhs_affine = False
    rhs_affine = False
    with scopes.dpp_scope():
        lhs_affine = lhs.is_affine()
        rhs_affine = rhs.is_affine()
    assert lhs_affine or rhs_affine

    if lhs_affine:
        t = Variable(rhs.shape)
        return lhs @ t, [t == rhs]
    else:
        t = Variable(lhs.shape)
        return t @ rhs, [t == lhs]



"""
Copyright 2013 Steven Diamond

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import Tuple

import numpy as np
import scipy.sparse as sp
from numpy.lib.array_utils import normalize_axis_index, normalize_axis_tuple

from cvxpy.atoms.atom import Atom


def normalize_axis(
    axis: int | tuple[int, ...], ndim: int, reduce_all_to_none: bool = True
) -> None | int | tuple[int, ...]:
    """Normalize an axis argument to a canonical form.

    - Negative indices become positive.
    - Single-element tuples become an int.
    - If all axes are listed and *reduce_all_to_none* is True, returns None.
    """
    axes = normalize_axis_tuple(axis, ndim)
    if reduce_all_to_none and len(axes) == ndim:
        return None
    elif len(axes) == 1:
        return axes[0]
    else:
        return axes


class AxisAtom(Atom):
    """
    An abstract base class for atoms that can be applied along an axis.
    """

    # Whether reducing over all axes is equivalent to axis=None.
    # True for reduction atoms (sum, max, min, etc.).
    # False for cumulative atoms (cumsum, cummax, cumprod) that preserve shape.
    _reduce_all_axes_to_none = True

    def __init__(
        self, expr, axis: None | int | tuple[int, ...] = None, keepdims: bool = False
    ) -> None:
        self.axis = axis
        self.keepdims = keepdims
        super(AxisAtom, self).__init__(expr)
        # Normalize axis after init so self.args is available.
        if self.axis is not None:
            ndim = len(self.args[0].shape)
            if ndim > 0:
                self.axis = normalize_axis(
                    self.axis, ndim, self._reduce_all_axes_to_none
                )

    def shape_from_args(self) -> Tuple[int, ...]:
        """
        Returns the shape of the atom after applying a function along an axis.
        Handles negative axis inputs by normalizing them to positive indices.
        """
        shape = list(self.args[0].shape)
        ndim = len(shape)
        if self.axis is None:
            return (1,) * len(shape) if self.keepdims else ()
        elif isinstance(self.axis, int):
            # Normalize negative axis
            axis = normalize_axis_index(self.axis, ndim)
            if self.keepdims:
                shape[axis] = 1
            else:
                shape = shape[:axis] + shape[axis+1:]
        else:
            # Normalize each axis in the list
            axes =  normalize_axis_tuple(self.axis, ndim)
            if self.keepdims:
                for axis in axes:
                    shape[axis] = 1
            else:
                shape[:] = [shape[i] for i in range(len(shape)) if i not in axes]
        return tuple(shape)

    def get_data(self):
        """
        Returns the axes and the keepdims parameter.
        """
        return [self.axis, self.keepdims]

    def validate_arguments(self) -> None:
        """
        Checks that each axis is within the valid range.
        """
        if self.axis is not None:
            axes = [self.axis] if isinstance(self.axis, int) else self.axis
            dim = self.args[0].ndim
            _ = normalize_axis_tuple(axes, dim)
        super(AxisAtom, self).validate_arguments()

    def _axis_grad(self, values) -> list[sp.csc_array] | None:
        """
        Gives the (sub/super)gradient of the atom w.r.t. each argument.

        Matrix expressions are vectorized, so the gradient is a matrix.
        Takes axis into account. Works for any number of dimensions.

        CVXPY convention: grad[i, j] = d(output_flat_F[j]) / d(input_flat_F[i])

        Args:
            values: A list of numeric values for the arguments.

        Returns:
            A list of SciPy CSC sparse matrices or None.
        """
        if self.axis is None or self.args[0].ndim < 2:
            value = np.reshape(values[0].T, (self.args[0].size, 1))
            D = self._column_grad(value)
            if D is not None:
                D = sp.csc_array(D)
            return [D]

        input_shape = self.args[0].shape
        ndim = len(input_shape)

        # Normalize axis to tuple
        axis = self.axis
        axes = (axis,) if isinstance(axis, int) else tuple(axis)
        keep = [i for i in range(ndim) if i not in axes]

        reduce_dims = [input_shape[a] for a in axes]
        reduce_size = int(np.prod(reduce_dims))
        output_shape = tuple(input_shape[i] for i in keep)
        input_size = int(np.prod(input_shape))
        output_size = max(1, int(np.prod(output_shape)))

        # F-order strides: stride[k] = prod(input_shape[:k])
        f_strides = np.ones(ndim, dtype=int)
        for k in range(1, ndim):
            f_strides[k] = f_strides[k-1] * input_shape[k-1]

        # Flat input in F-order
        flat_input = values[0].ravel(order='F')

        # All output multi-indices in F-order
        if len(output_shape) == 0:
            out_multis = np.zeros((0, 1), dtype=int)
        else:
            out_multis = np.array(
                np.unravel_index(np.arange(output_size), output_shape, order='F')
            )  # shape: (len(keep), output_size)

        # All reduce-axis multi-indices
        reduce_multis = np.array(
            np.unravel_index(np.arange(reduce_size), reduce_dims)
        ).T  # shape: (reduce_size, len(axes))

        all_rows = []
        all_cols = []
        all_data = []

        for j in range(output_size):
            om = out_multis[:, j]

            # Build input multi-indices: fix keep axes, vary reduce axes
            in_multis = np.zeros((reduce_size, ndim), dtype=int)
            for idx, k in enumerate(keep):
                in_multis[:, k] = om[idx]
            for idx, a in enumerate(axes):
                in_multis[:, a] = reduce_multis[:, idx]

            # Compute flat F-order indices for this fiber
            fiber_indices = in_multis @ f_strides
            fiber_values = flat_input[fiber_indices]

            d = self._column_grad(fiber_values.reshape(-1, 1))
            if d is None:
                return [None]
            d = np.asarray(d).flatten()

            all_rows.append(fiber_indices)
            all_cols.append(np.full(reduce_size, j, dtype=int))
            all_data.append(d)

        rows = np.concatenate(all_rows)
        cols = np.concatenate(all_cols)
        data = np.concatenate(all_data)
        D = sp.csc_array((data, (rows, cols)), shape=(input_size, output_size))
        return [D]

    def _column_grad(self, value):
        """
        Gives the (sub/super)gradient of the atom w.r.t. a column argument.

        Matrix expressions are vectorized, so the gradient is a matrix.

        Args:
            value: A numeric value for a column.

        Returns:
            A SciPy sparse matrix or None.
        """
        raise NotImplementedError()



"""
Copyright 2022, the CVXPY authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from typing import List, Tuple

import numpy as np
from scipy import linalg as LA
from scipy.special import entr

from cvxpy.atoms.atom import Atom
from cvxpy.constraints.constraint import Constraint


class von_neumann_entr(Atom):
    """
    Represents the von Neumann Entropy of the positive-definite matrix :math:`X,`

        .. math::
            -\\operatorname{tr}(X \\log X).

    Mathematically, this is equivalent to

        .. math::
            \\texttt{von_neumann_entr}(X) = -\\textstyle\\sum_{i=1}^n \\lambda_i \\log \\lambda_i

    where :math:`\\lambda_{i}` are the eigenvalues of :math:`X.`

    Parameters
    ----------
    X : Expression or numeric
        A PSD matrix

    quad_approx : Tuple[int,...]
        This is either an empty tuple (default) or a 2-tuple. If a 2-tuple, then this atom
        is approximately canonicalized. The approximations replace ExpCone constraints with SOC
        constraints based on a quadrature scheme from https://arxiv.org/abs/1705.00812;
        quad_approx[0] is the number of quadrature nodes and quad_approx[1] is the number of
        scaling points in the quadrature scheme.

    Notes
    -----
    This function does not assume :math:`\\operatorname{tr}(X)=1,` which would be required
    for most uses of this function in the context of quantum information theory. 
    """

    def __init__(self, X, quad_approx: Tuple[int, ...] = ()) -> None:
        self.quad_approx = quad_approx
        super(von_neumann_entr, self).__init__(X)

    def numeric(self, values):
        N = values[0]
        if hasattr(N, 'value'):
            N = N.value  # assume this is an ndarray
        w = LA.eigvalsh(N)
        val = np.sum(entr(w))
        return val

    def validate_arguments(self) -> None:
        """Verify that the argument is Hermitian."""
        if not self.args[0].is_hermitian():
            raise ValueError(
                f"""
                The argument {self.args[0].name()} to von_neumann_entr must be a Hermitian matrix.
                If you know for a fact that the input is Hermitian, wrap it with the hermitian_wrap
                atom before calling von_neumann_entr.
                """
            )
        
    def sign_from_args(self) -> Tuple[bool, bool]:
        """Returns sign (is positive, is negative) of the expression.
        """
        return (False, False)

    def is_atom_convex(self) -> bool:
        """Is the atom convex?
        """
        return False

    def shape_from_args(self) -> Tuple[int, ...]:
        """Returns the shape of the expression.
        """
        return tuple()

    def is_atom_concave(self) -> bool:
        """Is the atom concave?
        """
        return True

    def is_incr(self, idx) -> bool:
        """Is the composition non-decreasing in argument idx?
        """
        return False

    def is_decr(self, idx) -> bool:
        """Is the composition non-increasing in argument idx?
        """
        return False

    def get_data(self):
        return [self.quad_approx]

    def _grad(self, values):
        """Gives the (sub/super)gradient of the atom w.r.t. each argument.

        Matrix expressions are vectorized, so the gradient is a matrix.

        Args:
            values: A list of numeric values for the arguments.

        Returns:
            A list of SciPy CSC sparse matrices or None.
        """
        # N = values[0]
        # L = cholesky(N)
        # derivative = 2*(L + L * logm(np.dot(L.T, L)))
        # TODO: have to wrap derivative around scipy CSC sparse matrices
        #  compare to log_det atom.
        raise NotImplementedError()

    def _domain(self) -> List[Constraint]:
        """Returns constraints describing the domain of the node.
        """
        return [self.args[0] >> 0]



"""
Copyright 2018 Akshay Agrawal

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import Tuple

import numpy as np

import cvxpy.interface as intf
from cvxpy.atoms.affine.hstack import hstack
from cvxpy.atoms.axis_atom import AxisAtom


class Prod(AxisAtom):
    """Multiply the entries of an expression.

    The semantics of this atom are the same as np.prod.

    This atom is log-log affine, but it is neither convex nor concave.

    Parameters
    ----------
    expr : Expression
        The expression to multiply the entries of.
    axis : int
        The axis along which to sum.
    keepdims : bool
        Whether to drop dimensions after summing.
    """

    def __init__(self, expr, axis: None | int | tuple[int, ...] = None,
                 keepdims: bool = False) -> None:
        super(Prod, self).__init__(expr, axis=axis, keepdims=keepdims)

    def sign_from_args(self) -> Tuple[bool, bool]:
        """Returns sign (is positive, is negative) of the expression.
        """
        if self.args[0].is_nonneg():
            return (True, False)
        return (False, False)

    def is_atom_convex(self) -> bool:
        """Is the atom convex?
        """
        return False

    def is_atom_concave(self) -> bool:
        """Is the atom concave?
        """
        return False

    def is_atom_log_log_convex(self) -> bool:
        """Is the atom log-log convex?
        """
        return True

    def is_atom_log_log_concave(self) -> bool:
        """Is the atom log-log concave?
        """
        return True

    def is_incr(self, idx) -> bool:
        """Is the composition non-decreasing in argument idx?
        """
        return self.args[0].is_nonneg()

    def is_decr(self, idx) -> bool:
        """Is the composition non-increasing in argument idx?
        """
        return False

    def numeric(self, values):
        """Takes the product of the entries of value.
        """
        if intf.is_sparse(values[0]):
            sp_mat = values[0]
            if self.axis is None:
                if sp_mat.nnz == sp_mat.shape[0] * sp_mat.shape[1]:
                    data = sp_mat.data
                else:
                    data = np.zeros(1, dtype=sp_mat.dtype)
                result = np.prod(data)
            elif self.axis in [0, 1]:
                # The following snippet is taken from stackoverflow.
                # https://stackoverflow.com/questions/44320865/
                # can replace private _getnnz for scipy 1.15+ with count_nonzero
                mask = sp_mat._getnnz(axis=self.axis) == sp_mat.shape[self.axis]
                result = np.zeros(sp_mat.shape[1-self.axis], dtype=sp_mat.dtype)
                data = sp_mat[:, mask] if self.axis == 0 else sp_mat[mask, :]
                result[mask] = np.prod(data.toarray(), axis=self.axis)
                if self.keepdims:
                    result = np.expand_dims(result, self.axis)
            else:
                raise UserWarning("cp.prod does not support axis > 1 for sparse matrices.")
        else:
            result = np.prod(values[0], axis=self.axis, keepdims=self.keepdims)
        return result

    def _column_grad(self, value):
        """Gives the (sub/super)gradient of the atom w.r.t. a column argument.

        Matrix expressions are vectorized, so the gradient is a matrix.

        Args:
            value: A numeric value for a column.

        Returns:
            A NumPy ndarray or None.
        """
        return np.prod(value) / value

    def _grad(self, values):
        """Gives the (sub/super)gradient of the atom w.r.t. each argument.

        Matrix expressions are vectorized, so the gradient is a matrix.

        Args:
            values: A list of numeric values for the arguments.

        Returns:
            A list of SciPy CSC sparse matrices or None.
        """
        return self._axis_grad(values)


def prod(expr, axis=None, keepdims: bool = False) -> Prod:
    """Multiply the entries of an expression.

    The semantics of this atom are the same as np.prod.

    This atom is log-log affine, but it is neither convex nor concave.

    Parameters
    ----------
    expr : Expression or list[Expression, Numeric]
        The expression to multiply the entries of, or a list of Expressions
        and numeric types.
    axis : int
        The axis along which to take the product; ignored if `expr` is a list.
    keepdims : bool
        Whether to drop dimensions after taking the product; ignored if `expr`
        is a list.
    """
    if isinstance(expr, list):
        return Prod(hstack(expr))
    else:
        return Prod(expr, axis, keepdims)



"""
Copyright 2018 Akshay Agrawal

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import Tuple

import numpy as np
import scipy.sparse as sp

from cvxpy.atoms.affine.binary_operators import multiply
from cvxpy.atoms.atom import Atom


def diff_pos(x, y):
    r"""The difference :math:`x - y` with domain `\{x, y : x > y > 0\}`.

    This atom is log-log concave.

    Parameters
    ----------
    x : :class:`~cvxpy.expressions.expression.Expression`
        An Expression.
    y : :class:`~cvxpy.expressions.expression.Expression`
        An Expression.
    """
    return multiply(x, one_minus_pos(y/x))


class one_minus_pos(Atom):
    r"""The difference :math:`1 - x` with domain `\{x : 0 < x < 1\}`.

    This atom is log-log concave.

    Parameters
    ----------
    x : :class:`~cvxpy.expressions.expression.Expression`
        An Expression.
    """
    def __init__(self, x) -> None:
        super(one_minus_pos, self).__init__(x)
        self.args[0] = x
        self._ones = np.ones(self.args[0].shape)

    def numeric(self, values):
        return self._ones - values[0]

    def _grad(self, values):
        del values
        rows = self.args[0].size
        return [sp.dia_array((-np.ones(rows), [0]), shape=(rows, rows)).tocsc()]

    def name(self) -> str:
        return f"{self.__class__.__name__}({self.args[0]})"

    def format_labeled(self) -> str:
        if self._label is not None:
            return self._label
        return f"{self.__class__.__name__}({self.args[0].format_labeled()})"

    def shape_from_args(self) -> Tuple[int, ...]:
        """Returns the (row, col) shape of the expression.
        """
        return self.args[0].shape

    def sign_from_args(self) -> Tuple[bool, bool]:
        """Returns sign (is positive, is negative) of the expression.
        """
        return (True, False)

    def is_atom_convex(self) -> bool:
        """Is the atom convex?
        """
        return False

    def is_atom_concave(self) -> bool:
        """Is the atom concave?
        """
        return False

    def is_atom_log_log_convex(self) -> bool:
        """Is the atom log-log convex?
        """
        return False

    def is_atom_log_log_concave(self) -> bool:
        """Is the atom log-log concave?
        """
        return True

    def is_incr(self, idx) -> bool:
        """Is the composition non-decreasing in argument idx?
        """
        return False

    def is_decr(self, idx) -> bool:
        """Is the composition non-increasing in argument idx?
        """
        return True



import numpy as np

import cvxpy as cp


def test_clarabel_warm_start_sparsity_change():
    """
    Test that Clarabel gracefully falls back to new_solver() if sparsity changes
    during a warm start, preventing a Data formatting error. (Issue #2800)
    """
    x = cp.Variable(2)
    p = cp.Parameter(2)
    
    # Using sum so dropping a parameter to 0 remains mathematically feasible
    prob = cp.Problem(cp.Minimize(cp.sum_squares(x)), [cp.sum(cp.multiply(p, x)) >= 1])
    
    # Solve with fully dense parameter
    p.value = np.array([1.0, 1.0])
    prob.solve(solver=cp.CLARABEL, warm_start=True)
    
    # Introduce a zero to change the sparsity pattern of the data passed to Clarabel
    p.value = np.array([1.0, 0.0])
    
    # With the fix in PR #3225, this will gracefully fall back to a new solver
    # instead of crashing with a "Data formatting error"
    prob.solve(solver=cp.CLARABEL, warm_start=True)
    
    assert prob.status == cp.OPTIMAL



"""
Copyright 2024, the CVXPY authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import os
import subprocess

# IMPORTANT NOTE
#
#   Our versioning infrastructure is adapted from that used in SciPy v 1.9.0.
#   Specifically, much of this content came from
#   https://github.com/scipy/scipy/blob/91faf1ed4c3e83afe5009ffb7a9d18eab8dae683/tools/version_utils.py
#   It's possible that our adaptation has unnecessary complexity.
#   For example, SciPy might have certain contingencies in place for backwards
#   compatibilities that CVXPY does not guarantee.
#
#   Some comments in the SciPy source provide justification for individual code
#   snippets. We have mostly left those comments in-place, but we sometimes preface
#   them with the following remark:
#      "The comment below is from the SciPy code which we repurposed for cvxpy."
#

MAJOR = 1
MINOR = 8
MICRO = 2
IS_RELEASED = True
IS_RELEASE_BRANCH = True
VERSION = '%d.%d.%d' % (MAJOR, MINOR, MICRO)


# Return the git revision as a string
def git_version():
    def _minimal_ext_cmd(cmd):
        # construct minimal environment
        env = {}
        for k in ['SYSTEMROOT', 'PATH']:
            v = os.environ.get(k)
            if v is not None:
                env[k] = v
        # LANGUAGE is used on win32
        env['LANGUAGE'] = 'C'
        env['LANG'] = 'C'
        env['LC_ALL'] = 'C'
        out = subprocess.Popen(cmd, stdout=subprocess.PIPE, env=env).communicate()[0]
        return out

    try:
        out = _minimal_ext_cmd(['git', 'rev-parse', 'HEAD'])
        GIT_REVISION = out.strip().decode('ascii')[:7]

        # The comment below is from the SciPy code which we repurposed for cvxpy.
        #
        #   We need a version number that's regularly incrementing for newer commits,
        #   so the sort order in a wheelhouse of nightly builds is correct (see
        #   https://github.com/MacPython/scipy-wheels/issues/114). It should also be
        #   a reproducible version number, so don't rely on date/time but base it on
        #   commit history. This gives the commit count since the previous branch
        #   point from the current branch (assuming a full `git clone`, it may be
        #   less if `--depth` was used - commonly the default in CI):
        prev_version_tag = '^v{}.{}.0'.format(MAJOR, MINOR - 2)
        out = _minimal_ext_cmd(['git', 'rev-list', 'HEAD', prev_version_tag,
                                '--count'])
        COMMIT_COUNT = out.strip().decode('ascii')
        COMMIT_COUNT = '0' if not COMMIT_COUNT else COMMIT_COUNT
    except OSError:
        GIT_REVISION = "Unknown"
        COMMIT_COUNT = "Unknown"

    return GIT_REVISION, COMMIT_COUNT


def get_version_info():
    # The comment below is from the SciPy code which we adapted for cvxpy.
    #
    #   Adding the git rev number needs to be done inside
    #   write_version_py(), otherwise the import of cvxpy.version messes
    #   up the build under Python 3.
    FULLVERSION = VERSION
    if os.path.exists('.git'):
        GIT_REVISION, COMMIT_COUNT = git_version()
    elif os.path.exists('cvxpy/version.py'):
        # must be a source distribution, use existing version file
        # load it as a separate module to not load cvxpy/__init__.py
        import runpy
        ns = runpy.run_path('cvxpy/version.py')
        GIT_REVISION = ns['git_revision']
        COMMIT_COUNT = ns['git_revision']
    else:
        GIT_REVISION = "Unknown"
        COMMIT_COUNT = "Unknown"

    if not IS_RELEASED:
        FULLVERSION += '.dev0+' + COMMIT_COUNT + '.' + GIT_REVISION

    return FULLVERSION, GIT_REVISION, COMMIT_COUNT


def write_version_py(filename='cvxpy/version.py'):
    cnt = """
# THIS FILE IS GENERATED FROM CVXPY SETUP.PY
short_version = '%(version)s'
version = '%(version)s'
full_version = '%(full_version)s'
git_revision = '%(git_revision)s'
commit_count = '%(commit_count)s'
release = %(isrelease)s
if not release:
    version = full_version
"""
    FULLVERSION, GIT_REVISION, COMMIT_COUNT = get_version_info()

    a = open(filename, 'w')
    try:
        a.write(cnt % {'version': VERSION,
                       'full_version': FULLVERSION,
                       'git_revision': GIT_REVISION,
                       'commit_count': COMMIT_COUNT,
                       'isrelease': str(IS_RELEASED)})
    finally:
        a.close()



import numpy as np

from cvxpy import SOC, Variable, hstack
from cvxpy.constraints.exponential import ExpCone
from cvxpy.reductions.solvers.conic_solvers.scs_conif import (
    scs_psdvec_to_psdmat,
)
from cvxpy.utilities.solver_context import SolverInfo


def suppfunc_canon(expr, args, solver_context: SolverInfo | None = None):
    y = args[0].flatten(order="F")
    # ^ That's the user-supplied argument to the support function.
    parent = expr._parent
    A, b, K_sels = parent.conic_repr_of_set()
    # ^ That defines the set "X" associated with this support function.
    eta = Variable(shape=(b.size,))
    expr._eta = eta
    # ^ The main part of the duality trick for representing the epigraph
    # of this support function.
    n = A.shape[1]
    n0 = y.size
    if n > n0:
        # The description of the set "X" used in this support
        # function included n - n0 > 0 auxiliary variables.
        # We can pretend these variables were user-defined
        # by appending a suitable number of zeros to y.
        y_lift = hstack([y, np.zeros(shape=(n - n0,))])
    else:
        y_lift = y
    local_cons = [A.T @ eta + y_lift == 0]
    # now, the conic constraints on eta.
    #   nonneg, exp, soc, psd
    nonnegsel = K_sels["nonneg"]
    if nonnegsel.size > 0:
        temp_expr = eta[nonnegsel]
        local_cons.append(temp_expr >= 0)
    socsels = K_sels["soc"]
    for socsel in socsels:
        tempsca = eta[socsel[0]]
        tempvec = eta[socsel[1:]]
        soccon = SOC(tempsca, tempvec)
        local_cons.append(soccon)
    psdsels = K_sels["psd"]
    for psdsel in psdsels:
        curmat = scs_psdvec_to_psdmat(eta, psdsel)
        local_cons.append(curmat >> 0)
    expsel = K_sels["exp"]
    if expsel.size > 0:
        matexpsel = np.reshape(expsel, (-1, 3))
        curr_u = eta[matexpsel[:, 0]]
        curr_v = eta[matexpsel[:, 1]]
        curr_w = eta[matexpsel[:, 2]]
        # (curr_u, curr_v, curr_w) needs to belong to the dual
        # exponential cone, as used by the SCS solver. We map
        # this to a primal exponential cone as follows.
        ec = ExpCone(-curr_v, -curr_u, np.exp(1) * curr_w)
        local_cons.append(ec)
    epigraph = b @ eta
    return epigraph, local_cons
